#include "auto_pilot.h"
#include "route_planner.h"

#include <logging/logging.h>

#include <climits>

using Eigen::Vector2i;
using Eigen::Vector2d;
using Eigen::Vector3d;
using Eigen::Matrix3d;
using Eigen::Array2d;
using Eigen::Array3d;
using Eigen::Array2i;

AutoPilot::AutoPilot() {
    m_state = AutoPilotState::NAVIGATION;
    m_exp_search_timeout = 0;
    m_power_limit = 0.2;
    m_power = 0.0;
    m_destination_set = false;
    m_collision = true;

    m_next_frontier_search = 0;
    m_explorer_fail = false;

    memset(&m_front_segments, 0, sizeof(m_front_segments));
    memset(&m_side_segments, 0, sizeof(m_side_segments));
}

void AutoPilot::set_state(AutoPilotState state) {
    if (m_state == state)
        return;

    m_frontiers.clear();
    m_waypoints.clear();
    m_destination_set = false;

    m_state = state;
}

AutoPilotState AutoPilot::get_state() {
    return m_state;
}

Floormap *AutoPilot::get_floormap() const {
    return m_floormap;
}

void AutoPilot::set_floormap(Floormap *floormap) {
    m_floormap = floormap;

    m_exp_search.set_floormap(floormap);
}

void AutoPilot::set_power_limit(double power_limit) {
    m_power_limit = std::clamp(power_limit, 0.0, 1.0);
}

double AutoPilot::get_power_limit() const {
    return m_power_limit;
}

double AutoPilot::get_power() const {
    return m_power;
}

Vector2d AutoPilot::get_destination() const {
    if (!m_destination_set)
        warn("destination not set");

    return m_dest_rect.get_center();
}

void AutoPilot::set_destination(const Vector2d &destination) {
    if (m_state != AutoPilotState::NAVIGATION)
        return;

    m_dest_rect.dim = Vector2d(0.40, 0.40);
    m_dest_rect.pos = destination - m_dest_rect.dim / 2.0;

    m_destination_set = true;

    recalc_route();
}

void AutoPilot::clear_destination() {
    if (m_state != AutoPilotState::NAVIGATION)
        return;

    m_destination_set = false;
    m_waypoints.clear();
}

bool AutoPilot::has_destination() const {
    return m_destination_set;
}

void AutoPilot::push_lidar_data(const std::vector<LidarData> &lidar_data) {
    const double WIDTH = 0.22; // meter
    const double HEIGHT = 0.235; // meter
    const double THETA = 1.50;  // rad = 2 * arctan((HEIGHT / 2) / (WIDTH / 2))
    const double R = .30;       // meter
    const double DX = 0.05;     // meter
    const double DY = 0.05;     // meter

    memset(&m_side_segments, 0, sizeof(m_side_segments));
    memset(&m_front_segments, 0, sizeof(m_front_segments));

    for (const LidarData &datum: lidar_data) {
        double x = datum.dist * std::cos(datum.angle);
        double y = datum.dist * std::sin(datum.angle);

        // if infront determine the front segment index
        bool infront = (THETA / 2 > datum.angle || 2 * M_PI - THETA / 2 < datum.angle) &&
            (x >= WIDTH / 2 && datum.dist < R);
        if (infront) {
            if (y < -HEIGHT / 2)
                m_front_segments[0]++;
            else if (y < 0)
                m_front_segments[1]++;
            else if (y <= HEIGHT/ 2)
                m_front_segments[2]++;
            else /* if (y > HEIGHT / 2) */
                m_front_segments[3]++;
        }

        // check left segment
        if (!infront && (-WIDTH / 2 < x &&  x < WIDTH / 2 + DX) &&
                (-HEIGHT / 2 - DY < y && y < -HEIGHT / 2))
            m_side_segments[0]++;

        // check right segment
        if (!infront && (-WIDTH / 2 < x && x < WIDTH / 2 + DX) &&
                (HEIGHT / 2 < y && y < HEIGHT / 2 + DY))
            m_side_segments[1]++;
    }
}

void AutoPilot::get_waypoints(std::vector<Vector2d> &waypoints) const {
    waypoints = std::vector<Vector2d>(m_waypoints.begin(), m_waypoints.end());
}

void AutoPilot::get_frontiers(std::vector<Rect> &frontiers) const {
    frontiers = m_frontiers;
}

bool AutoPilot::get_collision() const {
    return m_collision;
}

void AutoPilot::set_collision(bool collision) {
    m_collision = collision;
}

void AutoPilot::tick() {
    assert(m_floormap != nullptr);

    // handle exploration search, restart every 5 seconds
    // after completion
    const Uint32 EXP_SEARCH_TIMEOUT = 5000;
    if (m_state == AutoPilotState::EXPLORING && m_exp_search.done()) {
        if (SDL_GetTicks() > m_exp_search_timeout) {
            std::vector<SearchResult> results;
            m_exp_search.get_results(results);

            m_frontiers.clear();
            for (const auto &result: results) {
                m_frontiers.push_back(result.rect);

                if (result.route.size() == 0)
                    continue;

                m_destination_set = true;
                m_dest_rect = result.rect;
                m_waypoints = result.route;
                recalc_route();
            }

            m_exp_search.start();
            m_exp_search_timeout = SDL_GetTicks() + EXP_SEARCH_TIMEOUT;
        }
    }
    else {
        // update the timeout while m_exp_search is running
        m_exp_search_timeout = SDL_GetTicks() + EXP_SEARCH_TIMEOUT;
    }

    // periodically recalc route
    if (SDL_GetTicks() > m_next_recalc_route)
        recalc_route();
}

void AutoPilot::get_speeds(int8_t &left, int8_t &right) {
    assert(m_floormap != nullptr);

    m_power = 0.0;
    left = 0;
    right = 0;

    Vector2d current_position;
    double current_heading;
    m_floormap->get_robot_position(current_position, current_heading);

    if (!route_good())
        return;

    // remove waypoints if we are close enough
    Vector2d error;
    while (!m_waypoints.empty()) {
        error = m_waypoints.front() - current_position;

        if (error.norm() > 0.05)
            break;

        m_waypoints.pop_front();
    }

    // stop when there are no waypoints and reset destination
    if (m_waypoints.empty()) {
        m_destination_set = false;
        return;
    }

    // stop when there is a collision
    if (m_collision)
        return;

    double angle_error = std::atan2(error[1], error[0]) - current_heading;
    if (angle_error < -M_PI)
        angle_error += 2 * M_PI;
    else if (angle_error > M_PI)
        angle_error -= 2 * M_PI;


    if (std::abs(angle_error) > 0.2) {
        m_power = 0.2;
        int sign = angle_error > 0 ? 1 : -1;
        left = 127 * m_power * sign;
        right = -127 * m_power * sign;
    }
    else {
        m_power = std::min(m_power_limit, 0.2 + 0.8 * error.norm());
        left = 127 * m_power;
        right = 127 * m_power;

        // do some immediate collision avoidance (based on the latest lidar data)
        double bias_side = (m_side_segments[1] - m_side_segments[0]) /
            std::max((double)(m_side_segments[0] + m_side_segments[1]), 1.0);

        double bias_ad = (m_front_segments[3] - m_front_segments[0]) /
            std::max((double)(m_front_segments[0] + m_front_segments[3]), 1.0);
        double bias_bc = (m_front_segments[2] - m_front_segments[1]) /
            std::max((double)(m_front_segments[1] + m_front_segments[2]), 1.0);

        double bias = 0.0; // negative -> left bias (turn left), pos -> right bias
        if (bias_bc > 0.0)
            bias = bias_bc;
        else
            bias = (bias_side + bias_ad) / 2.0;

        const double BIAS_STRENGTH = 0.4;

        left *= std::min(1.0, (1.0 - BIAS_STRENGTH * bias));
        right *= std::min(1.0, (1.0 + BIAS_STRENGTH * bias));
    }
}

bool AutoPilot::route_good() const {
    return !m_destination_set || !m_waypoints.empty();
}

bool AutoPilot::good() const {
    // return false when dst is set and there is no route
    if (!route_good())
        return false;

    if (m_collision)
        return false;

    return true;
}

bool AutoPilot::finished() const {
    return m_waypoints.empty() || !m_destination_set;
}

void AutoPilot::recalc_route() {
    m_next_recalc_route = SDL_GetTicks() + 2000;

    if (!m_destination_set)
        return;

    GridMap grid_map(2, 2);
    m_floormap->get_gm(grid_map);

    Vector2d robot_position;
    double robot_heading;
    m_floormap->get_robot_position(robot_position, robot_heading);

    plan_route(m_waypoints, grid_map, robot_position, m_dest_rect);
}

