#include "exploration_searcher.h"
#include "route_planner.h"

#include "logging/logging.h"

#include <algorithm>

using Eigen::Vector2d;
using Eigen::Vector2i;
using Eigen::Array2d;

namespace {

void floodfill(std::set<Vector2i, CompareVec> &reachable, const GridMap &gm, const Vector2i &start) {
    reachable.clear();
    std::set<Vector2i, CompareVec> queue = {start};

    const Vector2i neighbors[8] = {
        Vector2i(-1, -1), Vector2i(0, -1), Vector2i(1, -1),
        Vector2i(-1, 0),                   Vector2i(1, 0),
        Vector2i(-1, 1),  Vector2i(0, 1),  Vector2i(1, 1)
    };

    while (!queue.empty()) {
        Vector2i v = *queue.begin();
        queue.erase(v);
        reachable.insert(v);

        for (int i = 0; i < 8; i++) {
            Vector2i n = v + neighbors[i];

            if (gm.hit(n) > 0.5 || gm.miss(n) < 0.5)
                continue;

            if (reachable.find(n) != reachable.end())
                continue;

            queue.insert(n);
        }
    }
}

void frontier_search(std::vector<SearchResult> &results,
        const Vector2i &start, const GridMap &gm)
{
    std::set<Vector2i, CompareVec> unvisited = {start};
    std::set<Vector2i, CompareVec> visited;
    std::set<Vector2i, CompareVec> frontier_cells;

    const Vector2i neighbors[8] = {
        Vector2i(-1, -1), Vector2i(0, -1), Vector2i(1, -1),
        Vector2i(-1, 0), Vector2i(1, 0),
        Vector2i(-1, 1), Vector2i(0, 1), Vector2i(1, 1)
    };

    // use floodfill to search for the traversable area from the current
    // robot position (start).
    while (unvisited.size() > 0) {
        auto it = unvisited.begin();
        Vector2i v = *it;
        unvisited.erase(it);
        visited.insert(v);

        bool frontier_cell = false;
        for (size_t i = 0; i < 8; i++) {
            Vector2i n = v + neighbors[i];

            if (!gm.valid_coord(n))
                continue;

            if ((gm.hit(n) > 0.5 || gm.miss(n) > 0.5) && visited.find(n) == visited.end())
                unvisited.insert(n);

            if (gm.hit(v) < 0.5 && (gm.hit(n) < 0.5 && gm.miss(n) < 0.5))
                frontier_cell = true;
        }

        if (frontier_cell)
            frontier_cells.insert(v);
    }

    // search for neighboring frontier cells and add a frontier rect to frontiers
    // if there are enough (>5) cells in the vicinity.
    int n_frontiers = 0;
    while (frontier_cells.size() > 0) {
        auto it = frontier_cells.begin();
        std::set<Vector2i, CompareVec> unvisited2 = {*it};
        std::set<Vector2i, CompareVec> visited2;

        while (unvisited2.size() > 0) {
            auto it = unvisited2.begin();
            Vector2i v = *it;
            unvisited2.erase(it);
            frontier_cells.erase(v);
            visited2.insert(v);

            for (size_t i = 0; i < 8; i++) {
                Vector2i n = v + neighbors[i];

                // skip if neighbor is not in frontier_cells
                if (frontier_cells.find(n) == frontier_cells.end())
                    continue;

                // add to unvisited if it is not already visited
                if (visited2.find(n) == visited2.end())
                    unvisited2.insert(n);
            }
        }

        // if there are more than 20 frontier cells, add them to the
        // frontier set
        if (visited2.size() > 20) {
            Array2d min_v(1e10, 1e10);
            Array2d max_v(-1e10, -1e10);

            for (const Vector2i &vi: visited2) {
                Vector2d v = gm.convert_coord(vi);
                min_v = min_v.min(v.array());
                max_v = max_v.max(v.array());
            }

            results.push_back(
                {
                    Rect(min_v.matrix(), (max_v - min_v).matrix()),
                    SearchResultType::FRONTIER,
                    -1,
                    {},
                });
            n_frontiers++;
        }

    }
    info("found", n_frontiers, "frontiers");
}

void unvisited_search(std::vector<SearchResult> &results,
        const Vector2i &start, const GridMap &gm, const Map<int64_t> &hm) {

        std::set<Vector2i, CompareVec> reachable;
        floodfill(reachable, gm, start);

        // convert from gridmap integer coord to history map integer coords
        // this step also removes double coordinates
        std::set<Vector2i, CompareVec> reachable_hm;
        for (const Vector2i &v_gm: reachable) {
            Vector2d v = gm.convert_coord(v_gm);
            Vector2i v_hm = hm.convert_coord(v);
            reachable_hm.insert(v_hm);
        }

        // mark reachable areas in history map cell size
        std::set<Vector2i, CompareVec> nolocalmap_hm;
        for (const Vector2i &v_hm: reachable_hm) {
            if (hm[v_hm] < 0)
                nolocalmap_hm.insert(v_hm);
        }

        // clear m_frontiers, we will add our own frontiers
        double scale = hm.get_cell_size();
        for (const Vector2i &v_hm: nolocalmap_hm) {
            // create a rect for this vector
            Vector2d pos = hm.convert_coord(v_hm);
            Vector2d dim(scale, scale);
            Rect r(pos, dim);
            results.push_back({r, SearchResultType::UNVISITED, -1, {}});
        }
}

void old_age_search(std::vector<SearchResult> &results,
        const GridMap &gm, const Map<int64_t> &hm) {
    (void)results;
    (void)gm;
    (void)hm;
}

class Comparator {
    public:
        Comparator(const Vector2d &origin) : m_origin(origin) {};

        bool operator()(const SearchResult &l, const SearchResult &r) {
            if (static_cast<int>(l.type) != static_cast<int>(r.type))
                return static_cast<int>(l.type) < static_cast<int>(r.type);

            if (l.sort_value != r.sort_value)
                return l.sort_value < r.sort_value;

            double l_dist = (l.rect.get_center() - m_origin).norm();
            double r_dist = (r.rect.get_center() - m_origin).norm();
            return l_dist < r_dist;
        }

    private:
        Vector2d m_origin;
};

};

ExplorationSearcher::ExplorationSearcher() {
    m_quit = false;
    m_state = State::DONE;
    m_thread = std::make_unique<std::thread>(&ExplorationSearcher::worker, this);
}

ExplorationSearcher::~ExplorationSearcher() {
    m_quit = true;
    m_cond.notify_all();
    m_thread->join();
}

void ExplorationSearcher::set_floormap(Floormap *floormap) {
    m_lock.lock();
    m_floormap = floormap;
    m_lock.unlock();
}

Floormap *ExplorationSearcher::get_floormap() const {
    Floormap *ret;

    m_lock.lock();
    ret = m_floormap;
    m_lock.unlock();

    return ret;
}

void ExplorationSearcher::start() {
    m_lock.lock();
    if (m_state == State::DONE)
        m_state = State::START;
    m_lock.unlock();

    m_cond.notify_all();
}

bool ExplorationSearcher::done() const {
    bool ret;

    m_lock.lock();
    ret = m_state == State::DONE;
    m_lock.unlock();

    return ret;
}

void ExplorationSearcher::get_results(std::vector<SearchResult> &results) const {
    m_lock.lock();
    results = m_results;
    m_lock.unlock();
}

void ExplorationSearcher::worker() {
    std::unique_lock<std::mutex> lock(m_lock);

    while (!m_quit) {
        m_cond.wait(lock);

        if (m_state != State::START)
            continue;
        m_state = State::WORKING;

        if (m_floormap == nullptr) {
            m_results.clear();
            m_state = State::DONE;
            continue;
        }

        // obtain a copy of the current gridmap
        // this needs to be done while we have the lock
        Vector2d position;
        double heading;
        m_floormap->get_robot_position(position, heading);

        GridMap gm(2, 2);
        m_floormap->get_gm(gm);

        Map<int64_t> hm(2, 2);
        m_floormap->get_hm(hm);

        Vector2i start = gm.convert_coord(position);

        // unlock the mutex to and start processing
        std::vector<SearchResult> results;
        lock.unlock();

        frontier_search(results, start, gm);
        unvisited_search(results, start, gm, hm);
        old_age_search(results, gm, hm);

        std::sort(results.begin(), results.end(), Comparator(position));

        for (SearchResult &result: results) {
            if (plan_route(result.route, gm, position, result.rect))
                break; // if we find a valid route, stop calculating
        }

        lock.lock();

        m_results = results;
        m_state = State::DONE;
    }
}
