#include "floormap.h"

#include "logging/logging.h"

#include "slam/passes.h"

using Eigen::Vector2i;
using Eigen::Vector2d;
using Eigen::Vector3d;
using Eigen::Matrix3d;
using Eigen::Array3d;
using Eigen::Array2i;

Floormap::Floormap() :
    m_gm(2.0 / 100, 2048), m_hm(64.0 / 100, 512)
{
    m_timestamp_prev = 0;
    m_quit = false;

    m_robot_position = Vector2d(0.0, 0.0);
    m_heading_vel_exceeded = false;
    m_pr_vel_exceeded = false;

    m_hm.fill(-1);

    m_thread = std::make_unique<std::thread>(&Floormap::worker, this);
}

Floormap::~Floormap() {
    m_quit = true;
    m_cond.notify_one();

    m_thread->join();
}

void Floormap::push_lidar_data(uint32_t timestamp, const std::vector<LidarData> &data) {
    // create new point cloud
    PointCloud pc;
    for (const LidarData &ld: data) {
        double x = ld.dist * cos(ld.angle);
        double y = ld.dist * sin(ld.angle);

        pc.push_back({0, Vector2d(x, y)});
    }
    pc.calc_timestamps(m_timestamp_prev, timestamp);
    m_timestamp_prev = timestamp;

    // add the point cloud to the queue
    m_lock.lock();
    m_pc_queue.push(pc);
    m_lock.unlock();

    m_cond.notify_one();
}

void Floormap::push_imu_data(const IMUDatum &datum) {
    m_lock.lock();
    m_imu_data.push_back(datum);
    m_lock.unlock();

    m_cond.notify_one();
}

void Floormap::get_gm(GridMap &gm) {
    m_lock.lock();
    gm = m_gm;
    m_lock.unlock();
}

void Floormap::get_hm(Map<int64_t> &hm) {
    m_lock.lock();
    hm = m_hm;
    m_lock.unlock();
}

void Floormap::get_robot_position(Vector2d &vec, double &heading) {
    m_lock.lock();
    vec = m_robot_position;
    if (m_imu_data.size() > 0)
        heading = m_imu_data[m_imu_data.size() - 1].heading;
    else
        heading = 0.0;
    m_lock.unlock();
}

bool Floormap::get_heading_vel_exceeded() {
    bool ret;

    m_lock.lock();
    ret = m_heading_vel_exceeded;
    m_lock.unlock();

    return ret;
}

bool Floormap::get_pr_vel_exceeded() {
    bool ret;

    m_lock.lock();
    ret = m_pr_vel_exceeded;
    m_lock.unlock();

    return ret;
}

void Floormap::worker() {
    std::unique_lock<std::mutex> lock(m_lock);
    while (!m_quit) {
        m_cond.wait(lock);

        while (!m_pc_queue.empty()) {
            if (m_imu_data.size() == 0)
                break;

            const PointCloud &pc = m_pc_queue.front();
            uint64_t pc_end_timestamp = pc[pc.size() - 1].timestamp;
            uint64_t imu_end_timestamp = m_imu_data[m_imu_data.size() - 1].timestamp;
            if (pc_end_timestamp > imu_end_timestamp)
                break;

            // this is the new point cloud we are going to add
            PointCloud pc_copy = m_pc_queue.front();
            m_pc_queue.pop();

            pass_pc_clamp(pc_copy, 0.1, 4.0);
            m_imu_data.smooth_heading();
            pass_pc_imu_correct(pc_copy, m_imu_data);

            uint64_t t0 = pc_copy[0].timestamp;
            uint64_t t1 = pc_copy[pc_copy.size() - 1].timestamp;
            bool heading_vel_exceeded = pass_imu_max_heading_vel(m_imu_data, t0, t1) > 0.001;
            bool pr_vel_exceeded = pass_imu_max_pr_vel(m_imu_data, t0, t1) > 0.001;
            pc_copy.set_skip(heading_vel_exceeded || pr_vel_exceeded);

            GridMap gm_copy(2.0 / 100, 2);
            Vector2d robot_position;

            lock.unlock();
            m_graph.insert_pc(pc_copy);
            m_graph.get_gm(gm_copy);
            m_graph.get_robot_position(robot_position);

            m_heading_vel_exceeded = heading_vel_exceeded;
            m_pr_vel_exceeded = pr_vel_exceeded;
            lock.lock();

            m_gm = std::move(gm_copy);
            m_hm[robot_position] = t0;
            m_robot_position = robot_position;
        }
    }
}

