#include "iocontroller.h"
#include "steering_controller.h"
#include "imu.h"

#include <iostream>

const uint32_t HEARTBEAT_INTERVAL = 50; //ms

IOController::IOController(Program *const p, Network *const n) : program(p), network(n) {
    m_map_pan_dx = 0.0;
    m_map_pan_dy = 0.0;

    // set this to something invalid
    m_prev_motor_left = 0xffff;
    m_prev_motor_right = 0xffff;
}

void IOController::main_loop() {
    bool quit = false;
    SDL_Event e;

    SteeringController sc;

    uint32_t next_heartbeat = 0;

    while(!quit) {
        const uint32_t start = SDL_GetTicks();

        // SDL events
        while(SDL_PollEvent(&e)) {
            switch(e.type) {
                case SDL_QUIT:
                    network->send_pkt(NETWORK_TYPE_TERMINATE, nullptr, 0);
                    quit = true;
                    break;

                case SDL_CONTROLLERAXISMOTION:
                    if(e.caxis.axis == SDL_CONTROLLER_AXIS_LEFTX)
                        sc.set_steering(e.caxis.value / 32768.0);
                    else if(e.caxis.axis == SDL_CONTROLLER_AXIS_TRIGGERRIGHT)
                        sc.set_acceleration(e.caxis.value / 32767.0);
                    else if(e.caxis.axis == SDL_CONTROLLER_AXIS_TRIGGERLEFT)
                        sc.set_deceleration(e.caxis.value / 32767.0);
                    else if(e.caxis.axis == SDL_CONTROLLER_AXIS_RIGHTX)
                        m_map_pan_dx = e.caxis.value / 32768.0;
                    else if(e.caxis.axis == SDL_CONTROLLER_AXIS_RIGHTY)
                        m_map_pan_dy = e.caxis.value / 32768.0;
                    break;

                case SDL_CONTROLLERBUTTONDOWN:
                    if(e.cbutton.button == SDL_CONTROLLER_BUTTON_LEFTSHOULDER)
                        sc.set_limit(sc.get_limit() - .1);
                    else if(e.cbutton.button == SDL_CONTROLLER_BUTTON_RIGHTSHOULDER)
                        sc.set_limit(sc.get_limit() + .1);
                    else if (e.cbutton.button == SDL_CONTROLLER_BUTTON_BACK)
                        program->switch_view();
                    else if (e.cbutton.button == SDL_CONTROLLER_BUTTON_DPAD_UP)
                        program->minimap_size(true);
                    else if (e.cbutton.button == SDL_CONTROLLER_BUTTON_DPAD_DOWN)
                        program->minimap_size(false);
                    else if (e.cbutton.button == SDL_CONTROLLER_BUTTON_DPAD_RIGHT)
                        program->map_scale(true);
                    else if (e.cbutton.button == SDL_CONTROLLER_BUTTON_DPAD_LEFT)
                        program->map_scale(false);
                    else if (e.cbutton.button == SDL_CONTROLLER_BUTTON_RIGHTSTICK)
                        program->map_pan_focus_on_robot();
                    else if (e.cbutton.button == SDL_CONTROLLER_BUTTON_A &&
                            program->get_state() == ProgramState::MAIN_VIEW)
                        program->toggle_ap();
                    else if (e.cbutton.button == SDL_CONTROLLER_BUTTON_A &&
                            program->get_state() == ProgramState::MAP_VIEW)
                        program->set_destination();
                    else if (e.cbutton.button == SDL_CONTROLLER_BUTTON_B &&
                            program->get_state() == ProgramState::MAIN_VIEW)
                        program->switch_ap_state();
                    else if (e.cbutton.button == SDL_CONTROLLER_BUTTON_B &&
                            program->get_state() == ProgramState::MAP_VIEW)
                        program->reset_destination();
                    break;
            }
        }

        if (std::abs(m_map_pan_dx) > 0.1 || std::abs(m_map_pan_dy) > 0.1)
            program->map_pan(m_map_pan_dx / 100, m_map_pan_dy / 100);

        program->set_power(sc.get_power(), sc.get_limit());

        int8_t motors[2];
        if (program->get_ap_active())
            program->get_ap().get_speeds(motors[0], motors[1]);
        else
            sc.get_speeds(motors[0], motors[1]);
        if (motors[0] != m_prev_motor_left || motors[1] != m_prev_motor_right) {
            m_prev_motor_left = motors[0];
            m_prev_motor_right = motors[1];

            network->send_pkt(NETWORK_TYPE_MOTOR, motors, 2);
        }

        // send heartbeat event
        if (start > next_heartbeat) {
            network->send_pkt(NETWORK_TYPE_HEARTBEAT, nullptr, 0);
            next_heartbeat = SDL_GetTicks() + HEARTBEAT_INTERVAL;
        }

        // Network Events
        uint8_t pkt_type;
        char payload[NETWORK_MAX_PAYLOAD_SIZE];
        size_t payload_size = NETWORK_MAX_PAYLOAD_SIZE;
        uint32_t timestamp;
        while(network->recv_pkt(pkt_type, payload, payload_size)) {
            switch(pkt_type) {
                case NETWORK_TYPE_VIDEO_STREAM:
                    program->handle_camera_stream(payload, payload_size);
                    break;

                case NETWORK_TYPE_LIDAR_POINTS:
                    timestamp = ntohl(*(uint32_t *)payload);
                    program->handle_lidar_data(timestamp,
                            (uint16_t *)(payload + 4), (payload_size - 4) / 4);
                    break;

                case NETWORK_TYPE_BAT_LEVEL:
                    program->set_battery(static_cast<uint8_t>(payload[0]) / 255.0);
                    break;

                case NETWORK_TYPE_BAT_BROWNOUT:
                    program->brownout();
                    break;

                case NETWORK_TYPE_IMU:
                    for (int i = 0; i < 10; i++) {
                        IMU_Datum datum = imu_datum_unpack((uint8_t *)payload + 14 * i);
                        program->handle_imu_data(datum);
                    }
                    break;
            }
            payload_size = NETWORK_MAX_PAYLOAD_SIZE;
        }

        program->tick();

        // Rate limiting
        const uint32_t elapsed = SDL_GetTicks() - start;
        if(elapsed < SEC_PER_TICK)
            SDL_Delay(SEC_PER_TICK - elapsed);
    }
}
