#include "layout_widget.h"

#include "logging/logging.h"

#include <cassert>

LayoutWidget::LayoutWidget() {
    for (int i = 0; i < static_cast<int>(LayoutWidgetPlacement::N_PLACEMENTS); i++)
        m_children_sizes[i] = {0, 0};
}

void LayoutWidget::add_child(std::shared_ptr<Widget> widget, LayoutWidgetPlacement placement) {
    int widget_index = static_cast<int>(placement);
    if (m_children[widget_index]) {
        error("assigning multiple widgets to the same placement");
        return;
    }

    m_children[widget_index] = widget;
}

void LayoutWidget::remove_child(LayoutWidgetPlacement placement) {
    int index = static_cast<int>(placement);

    m_children[index].reset();
    m_children_sizes[index] = {0, 0};
}

Widget *LayoutWidget::get_child(LayoutWidgetPlacement placement) const {
    return m_children[static_cast<int>(placement)].get();
}

void LayoutWidget::request_size(WidgetSize &size, const WidgetSize &hint) {
    WidgetSize corner_hint = hint;
    corner_hint.width /= 2;
    corner_hint.height /= 2;

    for (int i = 0; i < static_cast<int>(LayoutWidgetPlacement::N_PLACEMENTS); i++) {
        WidgetSize child_hint;
        if (static_cast<int>(LayoutWidgetPlacement::CENTER) == i)
            child_hint = hint;
        else
            child_hint = corner_hint;

        if (m_children[i]) {
            m_children[i]->request_size(m_children_sizes[i], child_hint);
        }
        else {
            m_children_sizes[i].width = 0;
            m_children_sizes[i].height = 0;
        }
    }

    int min_width = m_children_sizes[static_cast<int>(LayoutWidgetPlacement::CENTER)].width;
    min_width = std::max(min_width,
            m_children_sizes[static_cast<int>(LayoutWidgetPlacement::TOP_LEFT)].width +
            m_children_sizes[static_cast<int>(LayoutWidgetPlacement::TOP_RIGHT)].width);
    min_width = std::max(min_width,
            m_children_sizes[static_cast<int>(LayoutWidgetPlacement::BOTTOM_LEFT)].width +
            m_children_sizes[static_cast<int>(LayoutWidgetPlacement::BOTTOM_RIGHT)].width);

    int min_height = m_children_sizes[static_cast<int>(LayoutWidgetPlacement::CENTER)].height;
    min_height = std::max(min_height,
            m_children_sizes[static_cast<int>(LayoutWidgetPlacement::TOP_LEFT)].height +
            m_children_sizes[static_cast<int>(LayoutWidgetPlacement::BOTTOM_LEFT)].height);
    min_height = std::max(min_height,
            m_children_sizes[static_cast<int>(LayoutWidgetPlacement::TOP_RIGHT)].height +
            m_children_sizes[static_cast<int>(LayoutWidgetPlacement::BOTTOM_RIGHT)].height);

    size.width = min_width;
    size.height = min_height;
}

void LayoutWidget::draw(SDL_Renderer *renderer, const WidgetDrawLocation &location) {
    assert(renderer != nullptr);

    int index;

    // center
    index = static_cast<int>(LayoutWidgetPlacement::CENTER);
    if (m_children[index])
        m_children[index]->draw(renderer, location);

    // top left
    index = static_cast<int>(LayoutWidgetPlacement::TOP_LEFT);
    if (m_children[index]) {
        WidgetDrawLocation top_left_location;
        top_left_location.size = m_children_sizes[index];
        top_left_location.x = location.x;
        top_left_location.y = location.y;
        m_children[index]->draw(renderer, top_left_location);
    }

    // top right
    index = static_cast<int>(LayoutWidgetPlacement::TOP_RIGHT);
    if (m_children[index]) {
        WidgetDrawLocation top_right_location;
        top_right_location.size = m_children_sizes[index];
        top_right_location.x = location.size.width - top_right_location.size.width;
        top_right_location.y = location.y;
        m_children[index]->draw(renderer, top_right_location);
    }

    // bottom left
    index = static_cast<int>(LayoutWidgetPlacement::BOTTOM_LEFT);
    if (m_children[index]) {
        WidgetDrawLocation bottom_left_location;
        bottom_left_location.size = m_children_sizes[index];
        bottom_left_location.x = location.x;
        bottom_left_location.y = location.size.height - bottom_left_location.size.height;
        m_children[index]->draw(renderer, bottom_left_location);
    }

    // bottom right
    index = static_cast<int>(LayoutWidgetPlacement::BOTTOM_RIGHT);
    if (m_children[index]) {
        WidgetDrawLocation bottom_right_location;
        bottom_right_location.size = m_children_sizes[index];
        bottom_right_location.x = location.size.width - bottom_right_location.size.width;
        bottom_right_location.y = location.size.height - bottom_right_location.size.height;
        m_children[index]->draw(renderer, bottom_right_location);
    }
}

