#include "arg_parser.h"
#include "iocontroller.h"
#include "program.h"

#include <logging/logging.h>
#include <network/network.h>

#include <SDL2/SDL.h>

int main(int argc, char *argv[]) {
    ArgParser arg_parser(argc, argv);
    if (!arg_parser.good())
        return EXIT_FAILURE;

    if(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER | SDL_INIT_GAMECONTROLLER | SDL_INIT_HAPTIC) < 0) {
        error("SDL_Init():", SDL_GetError());
        exit(EXIT_FAILURE);
    }

    SDL_Haptic *haptic = nullptr;
    bool has_joystick = false;
    info("Num joy sticks ", SDL_NumJoysticks());
    for(int i = 0; i < SDL_NumJoysticks(); i++) {
        SDL_GameController *controller = SDL_GameControllerOpen(i);
        if(controller == nullptr) {
            warn("Failed to open controller");
            continue;
        }
        has_joystick = true;

        // try to open haptic and init
        SDL_Joystick *joystick = SDL_GameControllerGetJoystick(controller);
        if (joystick == nullptr)
            continue;

        haptic = SDL_HapticOpenFromJoystick(joystick);
        if (haptic == nullptr)
            continue;

        if (SDL_HapticRumbleInit(haptic) != 0) {
            SDL_HapticClose(haptic);
            haptic = nullptr;
        }
    }
    if (!has_joystick) {
        error("At least one Joystick needs to be connected");
        return EXIT_FAILURE;
    }

    Network network(arg_parser.get_ip().c_str(), arg_parser.get_port(), false);
    Program program;
    program.set_haptic(haptic);
    IOController io_controller(&program, &network);

    io_controller.main_loop();

    if (arg_parser.get_save_raw()) {
        info("saving raw lidar and IMU data");

        int pc_counter = 0;
        for (const PointCloud &pc: program.get_point_clouds()) {
            char pc_filename[32];
            sprintf(pc_filename, "./%04d.pc", pc_counter++);
            pc.write(pc_filename);
        }

        program.get_imu_data().write("imu_data.imu");
    }

    SDL_Quit();

    return EXIT_SUCCESS;
}
