#include "box_widget.h"
#include "color_widget.h"
#include "fixed_widget.h"
#include "main_view.h"
#include "margin_widget.h"

#include <cassert>

namespace {

std::shared_ptr<BoxWidget> pack_indicators(std::shared_ptr<IndicatorWidget> indicators[4][4]) {
    std::shared_ptr<BoxWidget> box = std::make_shared<BoxWidget>(BoxWidgetOrientation::VERTICAL);

    for (size_t j = 0; j < 4; j++) {
        std::shared_ptr<BoxWidget> box1 =
            std::make_shared<BoxWidget>(BoxWidgetOrientation::HORIZONTAL);

        for (size_t i = 0; i < 4; i++) {
            if (indicators[i][j])
                box1->add_child(widget_cast(indicators[i][j]));
            else
                box1->add_child(widget_cast(std::make_shared<FixedWidget>(50, 50)));

            if (i < 3)
                box1->add_child(widget_cast(std::make_shared<FixedWidget>(30, 0)));
        }

        box->add_child(widget_cast(box1));
        box->add_child(widget_cast(std::make_shared<FixedWidget>(0, 10)));
    }

    return box;
}

std::shared_ptr<IndicatorWidget> load_indicator(const std::string &name, IndicatorWidgetColor color) {
    std::shared_ptr<IndicatorWidget> widget;

    widget = std::make_shared<IndicatorWidget>("indicators/" + name + ".png");
    widget->set_color(color);
    widget->blink();

    return widget;
}

}

MainView::MainView() {
    m_widget = std::make_shared<LayoutWidget>();

    m_streamer_widget = std::make_shared<StreamerWidget>();
    m_widget->add_child(widget_cast(m_streamer_widget), LayoutWidgetPlacement::CENTER);

    init_radar_widget();
    init_map_widget();
    init_gauges();
}

StreamerWidget &MainView::get_streamer_widget() {
    assert(m_streamer_widget != nullptr);

    return *m_streamer_widget;
}

MapWidget &MainView::get_map_widget() {
    assert(m_map_widget != nullptr);

    return *m_map_widget;
}

ScaleWidget &MainView::get_power_widget() {
    assert(m_power_widget != nullptr);

    return *m_power_widget;
}

RadarWidget &MainView::get_radar_widget() {
    assert(m_radar_widget != nullptr);

    return *m_radar_widget;
}

IndicatorWidget &MainView::get_ap_widget() {
    assert(m_ap_widget != nullptr);

    return *m_ap_widget;
}

IndicatorWidget &MainView::get_navigate_widget() {
    assert(m_navigate_widget != nullptr);

    return *m_navigate_widget;
}

IndicatorWidget &MainView::get_explorer_widget() {
    assert(m_explorer_widget != nullptr);

    return *m_explorer_widget;
}

IndicatorWidget &MainView::get_battery_widget() {
    assert(m_battery_widget != nullptr);

    return *m_battery_widget;
}

IndicatorWidget &MainView::get_brownout_widget() {
    assert(m_brownout_widget != nullptr);

    return *m_brownout_widget;
}

IndicatorWidget &MainView::get_camera_widget() {
    assert(m_camera_widget != nullptr);

    return *m_camera_widget;
}

IndicatorWidget &MainView::get_lidar_widget() {
    assert(m_lidar_widget != nullptr);

    return *m_lidar_widget;
}

IndicatorWidget &MainView::get_cpu_widget() {
    assert(m_cpu_widget != nullptr);

    return *m_cpu_widget;
}

IndicatorWidget &MainView::get_ram_widget() {
    assert(m_ram_widget != nullptr);

    return *m_ram_widget;
}

IndicatorWidget &MainView::get_collision_widget() {
    assert(m_collision_widget != nullptr);

    return *m_collision_widget;
}

IndicatorWidget &MainView::get_yaw_widget() {
    assert(m_yaw_widget != nullptr);

    return *m_yaw_widget;
}

IndicatorWidget &MainView::get_pr_widget() {
    assert(m_pr_widget != nullptr);

    return *m_pr_widget;
}

MainViewMapSize MainView::get_map_size() const {
    return m_map_size;
}

void MainView::set_map_size(MainViewMapSize size) {
    if (size == MainViewMapSize::SMALL)
        m_map_fixed_widget->set_size(300, 200);
    else if (size == MainViewMapSize::NORMAL)
        m_map_fixed_widget->set_size(400, 300);
    else if (size == MainViewMapSize::LARGE)
        m_map_fixed_widget->set_size(500, 400);
    else if (size == MainViewMapSize::HUGE)
        m_map_fixed_widget->set_size(800, 800);

    m_map_size = size;
}

Widget *MainView::get_widget() {
    return widget_cast(m_widget).get();
}

void MainView::init_radar_widget() {
    m_radar_widget = std::make_shared<RadarWidget>(35, 0.24);

    std::shared_ptr<FixedWidget> fixed = std::make_shared<FixedWidget>(300, 300);
    fixed->add_child(widget_cast(m_radar_widget));

    std::shared_ptr<MarginWidget> margin = std::make_shared<MarginWidget>();
    margin->add_child(widget_cast(fixed));

    m_widget->add_child(widget_cast(margin), LayoutWidgetPlacement::TOP_RIGHT);
}

void MainView::init_map_widget() {
    m_map_widget = std::make_shared<MapWidget>();
    m_map_widget->set_follow_robot(true);

    m_map_fixed_widget = std::make_shared<FixedWidget>(1, 1);
    m_map_fixed_widget->add_child(widget_cast(m_map_widget));
    set_map_size(MainViewMapSize::NORMAL);

    std::shared_ptr<MarginWidget> margin = std::make_shared<MarginWidget>();
    margin->add_child(widget_cast(m_map_fixed_widget));

    m_widget->add_child(widget_cast(margin), LayoutWidgetPlacement::BOTTOM_LEFT);
}

void MainView::init_gauges() {
    // load the inducator widgets
    m_ap_widget         = load_indicator("ap", IndicatorWidgetColor::RED);
    m_navigate_widget   = load_indicator("navigate", IndicatorWidgetColor::GREEN);
    m_explorer_widget   = load_indicator("explorer", IndicatorWidgetColor::GREEN);
    m_speedset_widget   = load_indicator("speedset", IndicatorWidgetColor::GREEN);
    m_battery_widget    = load_indicator("battery", IndicatorWidgetColor::RED);
    m_brownout_widget   = load_indicator("brownout", IndicatorWidgetColor::ORANGE);
    m_camera_widget     = load_indicator("camera", IndicatorWidgetColor::RED);
    m_lidar_widget      = load_indicator("lidar", IndicatorWidgetColor::RED);
    m_cpu_widget        = load_indicator("cpu", IndicatorWidgetColor::GREEN);
    m_ram_widget        = load_indicator("ram", IndicatorWidgetColor::GREEN);
    m_collision_widget  = load_indicator("collision", IndicatorWidgetColor::ORANGE);
    m_yaw_widget        = load_indicator("yaw", IndicatorWidgetColor::ORANGE);
    m_pr_widget         = load_indicator("pr", IndicatorWidgetColor::ORANGE);

    std::shared_ptr<IndicatorWidget> indicators[4][4];
    indicators[0][0] = m_ap_widget;
    indicators[1][0] = m_navigate_widget;
    indicators[2][0] = m_explorer_widget;
    indicators[3][0] = m_speedset_widget;
    indicators[0][1] = m_cpu_widget;
    indicators[1][1] = m_ram_widget;
    indicators[0][2] = m_collision_widget;
    indicators[1][2] = m_yaw_widget;
    indicators[2][2] = m_pr_widget;
    indicators[0][3] = m_camera_widget;
    indicators[1][3] = m_lidar_widget;
    indicators[2][3] = m_brownout_widget;
    indicators[3][3] = m_battery_widget;
    std::shared_ptr<BoxWidget> box = pack_indicators(indicators);

    // the power gauge widget
    m_power_widget = std::make_shared<ScaleWidget>();

    std::shared_ptr<FixedWidget> fixed = std::make_shared<FixedWidget>(290, 30);
    fixed->add_child(widget_cast(m_power_widget));

    box->add_child(widget_cast(fixed));

    std::shared_ptr<MarginWidget> margin = std::make_shared<MarginWidget>();
    margin->add_child(widget_cast(box));

    m_widget->add_child(widget_cast(margin), LayoutWidgetPlacement::BOTTOM_RIGHT);
}

