#include "map_widget.h"

#include <logging/logging.h>

using namespace Eigen;

namespace {

// taken from:
// http://www.sunshine2k.de/coding/java/TriangleRasterization/TriangleRasterization.html

void fill_bottom(uint32_t *raw, int width, int height, int pitch, Vector2i v[3], uint32_t color) {
    double invslope1 = (double)(v[1][0] - v[0][0]) / (v[1][1] - v[0][1]);
    double invslope2 = (double)(v[2][0] - v[0][0]) / (v[2][1] - v[0][1]);

    double curx1 = v[0][0];
    double curx2 = v[0][0];

    for (int y = v[0][1]; y <= v[1][1]; y++) {
        for (int x = std::min(curx1, curx2); x <= std::max(curx1, curx2); x++) {
            if (x < 0 || x >= width || y < 0 || y >= height)
                continue;
            raw[x + y * pitch] = color;
        }
        curx1 += invslope1;
        curx2 += invslope2;
    }
}

void fill_top(uint32_t *raw, int width, int height, int pitch, Vector2i v[3], uint32_t color) {
    double invslope1 = (double)(v[2][0] - v[0][0]) / (v[2][1] - v[0][1]);
    double invslope2 = (double)(v[2][0] - v[1][0]) / (v[2][1] - v[1][1]);

    double curx1 = v[2][0];
    double curx2 = v[2][0];

    for (int y = v[2][1]; y > v[0][1]; y--) {
        for (int x = std::min(curx1, curx2); x <= std::max(curx1, curx2); x++) {
            if (x < 0 || x >= width || y < 0 || y >= height)
                continue;
            raw[x + y * pitch] = color;
        }
        curx1 -= invslope1;
        curx2 -= invslope2;
    }
}

void draw_triangle(uint32_t *raw, int width, int height, int pitch, Vector2i v[3], uint32_t color) {
    std::sort(v, v + 3, [](auto &v, auto &w) { return v[1] < w[1]; });

    if (v[1][1] == v[2][1]) {
        fill_bottom(raw, width, height, pitch, v, color);
    }
    else if (v[0][1] == v[1][1]) {
        fill_top(raw, width, height, pitch, v, color);
    }
    else {
        Vector2i v4(
            v[0][0] + ((double)(v[1][1] - v[0][1]) / (v[2][1] - v[0][1])) * (v[2][0] - v[0][0]),
            v[1][1]
        );

        Vector2i bottom[3] = {v[0], v[1], v4};
        fill_bottom(raw, width, height, pitch, bottom, color);

        Vector2i top[3] = {v[1], v4, v[2]};
        fill_top(raw, width, height, pitch, top, color);
    }
}

void fill_square(uint32_t *raw, int width, int height, int pitch, Vector2i v, uint32_t color) {
    for (int j = v[1] - 1; j <= v[1] + 1; j++) {
        for (int i = v[0] - 1; i <= v[0] + 1; i++) {
            if (i < 0 || i > width || j < 0 || j >= height)
                continue;

            raw[i + j * pitch] = color;
        }
    }
}

void draw_line(uint32_t *raw, int width, int height, int pitch,
        const Vector2i &v0, const Vector2i &v1, uint32_t color)
{
    Vector2i diff = v1 - v0;
    int n_steps = std::max(std::abs(diff[0]), std::abs(diff[1]));

    for (int step = 0; step <= n_steps; step++) {
        Vector2i v = v0 + (diff.cast<double>() / n_steps * step).cast<int>();

        if (v[0] < 0 || v[0] >= width || v[1] < 0 || v[1] >= height )
            continue;

        raw[v[0] + pitch * v[1]] = color;
    }
}

}

MapWidget::MapWidget(double update_freq) :
    TextureComponent(SDL_PIXELFORMAT_RGBA8888, SDL_TEXTUREACCESS_STREAMING)
{
    if (update_freq < 0) {
        warn("update_freq can not be negative");
        update_freq = 0.1;
    }

    m_update_freq = 1000 * update_freq;
    m_next_update = SDL_GetTicks();

    m_floormap = nullptr;

    m_show_destination = false;
    m_destination = Vector2d(0.0, 0.0);

    m_scale = 128;
    m_offset = Vector2d(0.0, 0.0);

    m_show_cursor = false;
    m_follow_robot = false;
    m_pan_timeout = SDL_GetTicks();
}

void MapWidget::set_floormap(Floormap *floormap) {
    m_floormap = floormap;
}

Floormap *MapWidget::get_floormap() const {
    return m_floormap;
}

void MapWidget::set_waypoints(const std::vector<Eigen::Vector2d> &waypoints) {
    m_waypoints = waypoints;
}

void MapWidget::set_frontiers(const std::vector<Rect> &frontiers) {
    m_frontiers = frontiers;
}

void MapWidget::set_destination(const Eigen::Vector2d &destination) {
    m_destination = destination;
}

void MapWidget::show_destination(bool state) {
    m_show_destination = state;
}

int MapWidget::get_scale() const {
    return m_scale;
}

void MapWidget::set_scale(int scale) {
    m_scale = std::clamp(scale, 32, 512);
}

Eigen::Vector2d MapWidget::get_offset() {
    return m_offset;
}

void MapWidget::set_offset(const Eigen::Vector2d &offset) {
    // if we are following the robot, return to it after 2000 msec
    m_pan_timeout = SDL_GetTicks() + 2000;

    m_offset = offset;
}

bool MapWidget::get_show_cursor() const {
    return m_show_cursor;
}

void MapWidget::set_show_cursor(bool show_cursor) {
    m_show_cursor = show_cursor;
}

bool MapWidget::get_follow_robot() {
    return m_follow_robot;
}

void MapWidget::set_follow_robot(bool follow_robot) {
    m_follow_robot = follow_robot;
}

void MapWidget::request_size(WidgetSize &size, const WidgetSize &hint) {
    size = hint;
}

void MapWidget::draw(SDL_Renderer *renderer, const WidgetDrawLocation &location) {
    if (m_follow_robot && SDL_GetTicks() > m_pan_timeout) {
        Vector2d position;
        double heading;
        m_floormap->get_robot_position(position, heading);

        m_offset = position;
    }

    if (SDL_GetTicks() > m_next_update)
        update_texture(renderer, location.size.width, location.size.height);

    // draw the map
    SDL_Rect dst_rect = {location.x, location.y, location.size.width, location.size.height};
    if (!m_texture) {
        SDL_SetRenderDrawColor(renderer, 0x00, 0x00, 0x00, 0xff);
        SDL_RenderFillRect(renderer, &dst_rect);
        return;
    }
    SDL_RenderCopy(renderer, m_texture, nullptr, &dst_rect);
}

void MapWidget::update_texture(SDL_Renderer *renderer, int width, int height) {
    if (m_floormap == nullptr) {
        return;
    }

    allocate_texture(renderer, width, height);
    if (m_texture == nullptr)
        return;

    GridMap gm(2.0 / 100, 2);
    m_floormap->get_gm(gm);

    uint32_t *texture_raw;
    int texture_pitch;
    if (SDL_LockTexture(m_texture, nullptr, (void **)&texture_raw, &texture_pitch) != 0) {
        error("SDL_LockTexture", SDL_GetError());
        return;
    }
    texture_pitch /= sizeof(texture_raw[0]);

    // draw the map
    for (int j = 0; j < m_texture_height; j++) {
        for (int i = 0; i < m_texture_width; i++) {
            Vector2d coord = tex_to_map(Vector2i(i, j));

            uint8_t color;
            if (gm.valid_coord(coord)) {
                color = gm.miss(coord) * 127;
                if (gm.hit(coord) > 0.75)
                    color = 0xff;
            }
            else {
                // checker board pattern
                int a = i / 10;
                int b = j / 10;
                color = (a % 2) == (b % 2) ? 0x7f : 0x00;
            }

            texture_raw[i +  j * texture_pitch] =
                (color << 24) | (color << 16) | (color << 8) | 0xff;
        }
    }

    // draw the robot trail
    std::vector<Vector2d> robot_positions;
    gm.get_robot_positions(robot_positions);
    for (Vector2d &robot: robot_positions) {
        Vector2i robot_tex = map_to_tex(robot);

        for (int w = -1; w < 2; w++) {
            for (int z = -1; z < 2; z++) {
                int i = robot_tex[0] + w;
                int j = robot_tex[1] + z;

                if (i < 0 || i > m_texture_width || j < 0 || j > m_texture_height)
                    continue;

                texture_raw[i + j * texture_pitch] = 0x0000ffff;
            }
        }
    }

    // draw the robot position and its heading
    Vector2d robot_position;
    double robot_heading;
    m_floormap->get_robot_position(robot_position, robot_heading);

    Vector2i robot = map_to_tex(robot_position);
    Vector2i verts[3] = {Vector2i(10, 0), Vector2i(-10, 5), Vector2i(-10, -5)};
    for (size_t i = 0; i < sizeof(verts) / sizeof(verts[0]); i++)
        verts[i] = (Rotation2D(robot_heading) * verts[i].cast<double>()).cast<int>() + robot;
    draw_triangle(texture_raw, m_texture_width, m_texture_height, texture_pitch, verts, 0x0000ffff);

    // draw the waypoint path
    for (size_t i = 0; i < m_waypoints.size(); i++) {
        Vector2i from;
        if (i == 0)
            from = map_to_tex(robot_position);
        else
            from = map_to_tex(m_waypoints[i - 1]);
        Vector2i to = map_to_tex(m_waypoints[i]);

        Vector2i diff = to - from;
        int n_steps = std::max(std::abs(diff[0]), std::abs(diff[1]));

        for (int step = 0; step <= n_steps; step++) {
            Vector2i v = from + (diff.cast<double>() / n_steps * step).cast<int>();

            if (v[0] < 0 || v[0] >= m_texture_width || v[1] < 0 || v[1] >= m_texture_height)
                continue;

            texture_raw[v[0] + texture_pitch * v[1]] = 0x00ff00ff;
        }
    }

    // draw the waypoints
    for (const Vector2d &vec: m_waypoints) {
        Vector2i tex_vec = map_to_tex(vec);

        fill_square(texture_raw, m_texture_width, m_texture_height, texture_pitch,
                tex_vec, 0x00ff00ff);
    }

    // draw frontiers boxes
    for (const Rect &f: m_frontiers) {
        Vector2i verts[4];
        verts[0] = map_to_tex(Vector2d(f.get_xmin(), f.get_ymin()));
        verts[1] = map_to_tex(Vector2d(f.get_xmin(), f.get_ymax()));
        verts[2] = map_to_tex(Vector2d(f.get_xmax(), f.get_ymax()));
        verts[3] = map_to_tex(Vector2d(f.get_xmax(), f.get_ymin()));

        draw_line(texture_raw, m_texture_width, m_texture_height, texture_pitch,
                verts[0], verts[1], 0x00ff00ff);
        draw_line(texture_raw, m_texture_width, m_texture_height, texture_pitch,
                verts[1], verts[2], 0x00ff00ff);
        draw_line(texture_raw, m_texture_width, m_texture_height, texture_pitch,
                verts[2], verts[3], 0x00ff00ff);
        draw_line(texture_raw, m_texture_width, m_texture_height, texture_pitch,
                verts[3], verts[0], 0x00ff00ff);
    }

    // draw the destination marker
    if (m_show_destination) {
        fill_square(texture_raw, m_texture_width, m_texture_height, texture_pitch,
                map_to_tex(m_destination), 0x00ff00ff);
    }

    // draw the red cursor
    if (m_show_cursor) {
        fill_square(texture_raw, m_texture_width, m_texture_height, texture_pitch,
                Vector2i(m_texture_width / 2, m_texture_height / 2), 0xff0000ff);
    }

    SDL_UnlockTexture(m_texture);

    m_next_update = SDL_GetTicks() + m_update_freq;
}

