#include "program.h"
#include "lidar_data.h"

#include <logging/logging.h>

#include <iostream>
#include <iomanip>
#include <vector>
#include <arpa/inet.h>

using namespace Eigen;

Program::Program() {
    m_haptic = nullptr;

    m_state = ProgramState::MAIN_VIEW;

    m_main_view.get_streamer_widget().set_streamer(&m_streamer);
    m_main_view.get_map_widget().set_floormap(&m_floormap);
    m_main_view.get_navigate_widget().turn_on();
    m_map_view.get_map_widget().set_floormap(&m_floormap);


    m_ap_active = false;
    m_auto_pilot.set_floormap(&m_floormap);

    m_prev_timestamp = 0;
}

SDL_Haptic *Program::get_haptic() const {
    return m_haptic;
}

void Program::set_haptic(SDL_Haptic *haptic) {
    m_haptic = haptic;
}

void Program::handle_camera_stream(const void *data, size_t size) {
    m_streamer.push_data(data, size);
}

void Program::handle_lidar_data(uint32_t timestamp, uint16_t *data, int n) {
    std::vector<LidarData> lidar_data;

    const double ANGLE_OFFSET = 0.0592 + M_PI;
    for(int i = 0; i < n; i++) {
        LidarData point;
        point.angle = 2 * M_PI * ntohs(data[2 * i]) / UINT16_MAX + ANGLE_OFFSET;  // radian
        point.dist = ntohs(data[2 * i + 1]) / 1000.0;  // meter

        point.angle -= 2 * M_PI * static_cast<int>(point.angle / (2 * M_PI));

        lidar_data.push_back(point);
    }

    m_main_view.get_radar_widget().push_lidar_data(lidar_data);
    m_floormap.push_lidar_data(timestamp, lidar_data);
    m_auto_pilot.push_lidar_data(lidar_data);

    PointCloud pc;
    for (const LidarData &datum: lidar_data) {
        double x = datum.dist * std::cos(datum.angle);
        double y = datum.dist * std::sin(datum.angle);

        PointCloudDatum pc_datum;
        pc_datum.timestamp = 0;
        pc_datum.data = Vector2d(x, y);
        pc.push_back(pc_datum);
    }
    pc.calc_timestamps(m_prev_timestamp, timestamp);
    m_prev_timestamp = timestamp;
    m_point_clouds.push_back(pc);

    // collision detection
    IndicatorWidget &widget = m_main_view.get_collision_widget();
    bool collision_close = false;
    bool collision = false;
    for (size_t i = 0; i < pc.size(); i++) {
        PointCloudDatum datum = pc[i];

        if (datum.data.norm() < 0.20)
            collision_close = true;

        double x = datum.data[0];
        double y = datum.data[1];

        const double ROBOT_SIZE = 0.24;
        if ((x > -ROBOT_SIZE / 2 && x < ROBOT_SIZE / 2) &&
            (y > -ROBOT_SIZE / 2 && y < ROBOT_SIZE / 2)) {
            collision = true;
        }
    }

    if (collision) {
        widget.turn_on();
        widget.set_color(IndicatorWidgetColor::RED);

        SDL_HapticRumblePlay(m_haptic, 0.3, 500);
    }
    else if (collision_close) {
        widget.turn_on();
        widget.set_color(IndicatorWidgetColor::ORANGE);

        SDL_HapticRumblePlay(m_haptic, 0.1, 500);
    }
    else {
        widget.turn_off();
    }

    m_auto_pilot.set_collision(collision);
}

void Program::handle_imu_data(const IMU_Datum &datum) {
    IMUDatum d;
    d.timestamp = datum.timestamp;
    d.heading   = 2.0 * M_PI * datum.heading / 360.0;
    d.pitch     = 2.0 * M_PI * datum.pitch   / 360.0;
    d.roll      = 2.0 * M_PI * datum.roll    / 360.0;

    m_floormap.push_imu_data(d);
    m_imu_data.push_back(d);
}

void Program::tick() {
    m_auto_pilot.tick();

    // update the waypoints on the map widgets
    std::vector<Vector2d> waypoints;
    m_auto_pilot.get_waypoints(waypoints);

    m_main_view.get_map_widget().set_waypoints(waypoints);
    m_map_view.get_map_widget().set_waypoints(waypoints);

    if (m_auto_pilot.has_destination()) {
        Vector2d dst = m_auto_pilot.get_destination();

        m_main_view.get_map_widget().set_destination(dst);
        m_main_view.get_map_widget().show_destination(true);

        m_map_view.get_map_widget().set_destination(dst);
        m_map_view.get_map_widget().show_destination(true);
    }
    else {
        m_main_view.get_map_widget().show_destination(false);
        m_map_view.get_map_widget().show_destination(false);
    }

    // set frontiers in map widgets
    std::vector<Rect> frontiers;
    m_auto_pilot.get_frontiers(frontiers);
    m_main_view.get_map_widget().set_frontiers(frontiers);
    m_map_view.get_map_widget().set_frontiers(frontiers);

    // set the color of the AP indicator
    if (m_auto_pilot.finished())
        m_main_view.get_ap_widget().set_color(IndicatorWidgetColor::ORANGE);
    else if (!m_auto_pilot.good())
        m_main_view.get_ap_widget().set_color(IndicatorWidgetColor::RED);
    else
        m_main_view.get_ap_widget().set_color(IndicatorWidgetColor::GREEN);

    // set the color of the nav indicator
    if (!m_auto_pilot.has_destination())
        m_main_view.get_navigate_widget().set_color(IndicatorWidgetColor::ORANGE);
    else if (m_auto_pilot.route_good())
        m_main_view.get_navigate_widget().set_color(IndicatorWidgetColor::GREEN);
    else
        m_main_view.get_navigate_widget().set_color(IndicatorWidgetColor::RED);

    // set blink the heading/pr exceeded indicator
    if (m_floormap.get_heading_vel_exceeded())
        m_main_view.get_yaw_widget().blink();
    if (m_floormap.get_pr_vel_exceeded())
        m_main_view.get_pr_widget().blink();

    // draw the currently selected view
    if (m_state == ProgramState::MAIN_VIEW)
        m_graphics.draw(m_main_view.get_widget());
    else
        m_graphics.draw(m_map_view.get_widget());
}

void Program::set_battery(double level) {
    IndicatorWidget &widget = m_main_view.get_battery_widget();
    widget.turn_on();
    if (level > .6)
        widget.set_color(IndicatorWidgetColor::GREEN);
    else if (level > .2)
        widget.set_color(IndicatorWidgetColor::ORANGE);
    else
        widget.set_color(IndicatorWidgetColor::RED);
}

void Program::brownout() {
    m_main_view.get_brownout_widget().blink();
}

void Program::set_power(double value, double limit) {
    ScaleWidget &widget = m_main_view.get_power_widget();

    // value and limit passed are the values obtained from
    // the steering controller. If the auto pilot is active,
    // override the value variable
    if (m_ap_active) {
        m_auto_pilot.set_power_limit(limit);
        value = m_auto_pilot.get_power();
    }

    widget.set_value(std::abs(value));
    widget.set_limit(limit);
}

ProgramState Program::get_state() {
    return m_state;
}

void Program::switch_view() {
    if (m_state == ProgramState::MAIN_VIEW)
        m_state = ProgramState::MAP_VIEW;
    else
        m_state = ProgramState::MAIN_VIEW;
}

void Program::switch_ap_state() {
    AutoPilotState state = m_auto_pilot.get_state();
    if (state == AutoPilotState::NAVIGATION) {
        m_main_view.get_navigate_widget().turn_off();
        m_main_view.get_explorer_widget().turn_on();
        state = AutoPilotState::EXPLORING;
    }
    else /* if (state == AutoPilotState::EXPLORING) */ {
        m_main_view.get_navigate_widget().turn_on();
        m_main_view.get_explorer_widget().turn_off();
        state = AutoPilotState::NAVIGATION;
    }
    m_auto_pilot.set_state(state);
}

void Program::toggle_ap() {
    if (m_ap_active) {
        m_main_view.get_ap_widget().turn_off();
        m_ap_active = false;
    }
    else {
        m_main_view.get_ap_widget().turn_on();
        m_ap_active = true;
    }
}

bool Program::get_ap_active() {
    return m_ap_active;
}

AutoPilot &Program::get_ap() {
    return m_auto_pilot;
}

void Program::minimap_size(bool increase) {
    if (m_state != ProgramState::MAIN_VIEW)
        return;

    MainViewMapSize size = m_main_view.get_map_size();
    int new_size = static_cast<int>(size) + (increase ? 1 : -1);
    if (new_size < 0 || new_size >= static_cast<int>(MainViewMapSize::MAX))
        return;

    m_main_view.set_map_size(static_cast<MainViewMapSize>(new_size));
}

void Program::map_scale(bool increase) {
    int scale;

    if (m_state == ProgramState::MAIN_VIEW)
        scale = m_main_view.get_map_widget().get_scale();
    else
        scale = m_map_view.get_map_widget().get_scale();

    if (increase)
        scale *= 2;
    else
        scale /= 2;

    if (m_state == ProgramState::MAIN_VIEW)
        m_main_view.get_map_widget().set_scale(scale);
    else
        m_map_view.get_map_widget().set_scale(scale);
}

void Program::map_pan(double dx, double dy) {
    Vector2d offset;

    if (m_state == ProgramState::MAIN_VIEW)
        offset = m_main_view.get_map_widget().get_offset();
    else
        offset = m_map_view.get_map_widget().get_offset();

    offset += Vector2d(dx, dy);

    if (m_state == ProgramState::MAIN_VIEW)
        m_main_view.get_map_widget().set_offset(offset);
    else
        m_map_view.get_map_widget().set_offset(offset);
}

void Program::map_pan_focus_on_robot() {
    Vector2d position;
    double heading;
    m_floormap.get_robot_position(position, heading);

    if (m_state == ProgramState::MAIN_VIEW)
        m_main_view.get_map_widget().set_offset(position);
    else
        m_map_view.get_map_widget().set_offset(position);
}

void Program::set_destination() {
    Vector2d destination = m_map_view.get_map_widget().get_offset();

    m_auto_pilot.set_destination(destination);
    info("set destination at", destination[0], destination[1]);
}

void Program::reset_destination() {
    m_auto_pilot.clear_destination();
}

const IMUData &Program::get_imu_data() const {
    return m_imu_data;
}

const std::vector<PointCloud> &Program::get_point_clouds() const {
    return m_point_clouds;
}
