#ifndef __program_h__
#define __program_h__

#include "auto_pilot.h"
#include "floormap.h"
#include "graphics.h"
#include "imu.h"
#include "main_view.h"
#include "map_view.h"
#include "streamer.h"

#include <slam/imu_data.h>
#include <slam/point_cloud.h>

#include <SDL2/SDL.h>

#include <mutex>
#include <vector>

enum class ProgramState {
    MAIN_VIEW,
    MAP_VIEW
};

class Program {
    public:
        Program();

        Program(const Program &) = delete;
        Program &operator=(const Program &) = delete;

        SDL_Haptic *get_haptic() const;
        void set_haptic(SDL_Haptic *haptic);

        void tick();

        void handle_camera_stream(const void *data, size_t size);
        void handle_lidar_data(uint32_t timestamp, uint16_t *data, int n);
        void handle_imu_data(const IMU_Datum &datum);

        void set_battery(double level);
        void brownout();

        void set_power(double value, double limit);

        ProgramState get_state();
        void switch_view();
        void switch_ap_state();
        void toggle_ap();
        bool get_ap_active();
        AutoPilot &get_ap();
        void minimap_size(bool increase);
        void map_scale(bool increase);
        void map_pan(double dx, double dy);
        void map_pan_focus_on_robot();
        void set_destination();
        void reset_destination();

        const IMUData &get_imu_data() const;
        const std::vector<PointCloud> &get_point_clouds() const;

    private:
        SDL_Haptic     *m_haptic;

        ProgramState    m_state;
        MainView        m_main_view;
        MapView         m_map_view;

        bool            m_ap_active;
        AutoPilot       m_auto_pilot;

        Graphics        m_graphics;
        Streamer        m_streamer;
        Floormap        m_floormap;

        // used for saving the raw data
        IMUData                 m_imu_data;
        std::vector<PointCloud> m_point_clouds;
        uint32_t                m_prev_timestamp;
};

#endif  //#ifndef __program_h__
