#include "radar_widget.h"

#include <logging/logging.h>

RadarWidget::RadarWidget(double camera_angle, double robot_size) :
    TextureComponent(SDL_PIXELFORMAT_RGBA8888, SDL_TEXTUREACCESS_TARGET)
{
    m_camera_angle = camera_angle / 360.0 * 2 * M_PI;
    m_robot_size = robot_size;

    m_scale = 50.0;
}

void RadarWidget::set_scale(double scale) {
    if (scale <= 0.0) {
        error("scale must be larger than 0");
        scale = 1.0;
    }

    m_scale = scale;
}

double RadarWidget::get_scale() const {
    return m_scale;
}

void RadarWidget::push_lidar_data(const std::vector<LidarData> &data) {
    m_data = data;
}

void RadarWidget::request_size(WidgetSize &size, const WidgetSize &hint) {
    size = hint;
}

void RadarWidget::draw(SDL_Renderer *renderer, const WidgetDrawLocation &location) {
    allocate_texture(renderer, location.size.width, location.size.height);
    if (m_texture == nullptr)
        return;

    SDL_SetRenderTarget(renderer, m_texture);
    SDL_SetRenderDrawColor(renderer, 0x50, 0x50, 0x50, 0xff);
    SDL_RenderClear(renderer);

    // draw the lidar data
    SDL_SetRenderDrawColor(renderer, 0xff, 0xff, 0xff, 0xff);
    for(size_t i = 0; i < m_data.size(); i++) {
        size_t j = i + 1;
        if (j == m_data.size())
            j = 0;

        double angle_i = m_data[i].angle - M_PI / 2;
        double angle_j = m_data[j].angle - M_PI / 2;

        int x0 = (m_data[i].dist * cos(angle_i) * m_scale) + (location.size.width / 2);
        int y0 = (m_data[i].dist * sin(angle_i) * m_scale) + (location.size.height / 2);

        int x1 = (m_data[j].dist * cos(angle_j) * m_scale) + (location.size.width / 2);
        int y1 = (m_data[j].dist * sin(angle_j) * m_scale) + (location.size.height / 2);

        SDL_RenderDrawLine(renderer, x0, y0, x1, y1);
    }

    // draw camera lines
    SDL_SetRenderDrawColor(renderer, 0x00, 0xff, 0x00, 0xff);
    const double CAMERA_LINE_LENGTH = 1.0 * m_scale; // meter
    int x = (location.size.width / 2);
    int y = (location.size.height / 2);
    SDL_RenderDrawLine(renderer, x, y,
            x + CAMERA_LINE_LENGTH * cos(M_PI / 2 - m_camera_angle / 2),
            y - CAMERA_LINE_LENGTH * sin(M_PI / 2 - m_camera_angle / 2));
    SDL_RenderDrawLine(renderer, x, y,
            x + CAMERA_LINE_LENGTH * cos(M_PI / 2 + m_camera_angle / 2),
            y - CAMERA_LINE_LENGTH * sin(M_PI / 2 + m_camera_angle / 2));

    // draw the robot
    SDL_SetRenderDrawColor(renderer, 0xff, 0xff, 0xff, 0xff);
    SDL_Rect robot;
    robot.x = -m_scale * (m_robot_size / 2) + (location.size.width / 2);
    robot.y = -m_scale * (m_robot_size / 2) + (location.size.width / 2);
    robot.w = m_scale * m_robot_size;
    robot.h = m_scale * m_robot_size;
    SDL_RenderFillRect(renderer, &robot);

    SDL_SetRenderTarget(renderer, nullptr);

    SDL_Rect dst_rect = {location.x, location.y, location.size.width, location.size.height};
    SDL_RenderCopy(renderer, m_texture, nullptr, &dst_rect);
}

