#include "route_planner.h"

#include <logging/logging.h>

#include <map>
#include <vector>

using Eigen::Vector2d;
using Eigen::Vector2i;

namespace {

template <typename T>
int get_default(const std::map<Vector2i, T, CompareVec> &map, const Vector2i &vec, T def) {
    auto it = map.find(vec);
    if (it == map.end())
        return def;

    return it->second;
}

double dist_to_rect(const Rect &r, const Vector2d &v) {
    if (r.inside(v))
        return 0;

    if (r.get_xmin() < v[0] && v[0] < r.get_xmax()) {
        return std::min(std::abs(v[1] - r.get_ymin()), std::abs(v[1] - r.get_ymax()));
    }
    else if (r.get_ymin() < v[1] && v[1] < r.get_ymax()) {
        return std::min(std::abs(v[0] - r.get_xmin()), std::abs(v[0] - r.get_xmax()));
    }

    double corners[4];
    corners[0] = (Vector2d(r.get_xmin(), r.get_ymin()) - v).norm();
    corners[1] = (Vector2d(r.get_xmin(), r.get_ymax()) - v).norm();
    corners[2] = (Vector2d(r.get_xmax(), r.get_ymax()) - v).norm();
    corners[3] = (Vector2d(r.get_xmax(), r.get_ymin()) - v).norm();

    std::sort(&corners[0], &corners[4]);
    return corners[0];
}

bool valid_cell(const GridMap &gm, const Vector2i &coord, double r = 0.40) {
    double cell_size = gm.get_cell_size();
    int delta = std::ceil(r / cell_size) / 2;

    for (int y = -delta; y <= delta; y++) {
        for (int x = -delta; x <= delta; x++) {
            Vector2i v = coord + Vector2i(x, y);

            if (!gm.valid_coord(v))
                return false;

            if (gm.hit(v) > 0.5 || gm.miss(v) < 0.5)
                return false;
        }
    }

    return true;
}

bool find_valid_cell(const GridMap &gm, Vector2i &coord, double r = 0.40) {
    double cell_size = gm.get_cell_size();
    int delta = std::ceil(r / cell_size) / 2;

    bool valid_cell_found = false;
    double best_r = 1e10;
    Vector2i best_v = coord;

    for (int y = -delta; y <= delta; y++) {
        for (int x = -delta; x <= delta; x++) {
            Vector2i v = coord + Vector2i(x, y);

            if (!gm.valid_coord(v))
                continue;

            if (!valid_cell(gm, v))
                continue;

            double r = std::sqrt(x * x + y * y);
            if (r < best_r) {
                valid_cell_found = true;
                best_r = r;
                best_v = v;
            }
        }
    }

    coord = best_v;

    return valid_cell_found;
}

bool a_star(std::list<Vector2i> &path, const GridMap &gm,
        const Vector2i &from, const Rect &goal)
{
    path.clear();

    Vector2d center = goal.pos + goal.dim / 2.0;

    // use A star to calculate a path from the robot to the destinatino
    // add a waypoint for each cell (unneeded waypoints will later be pruned)
    // source: https://en.wikipedia.org/wiki/A*_search_algorithm
    std::set<Vector2i, CompareVec> open_set;
    std::map<Vector2i, Vector2i, CompareVec> came_from;
    std::map<Vector2i, double, CompareVec> g_score;
    std::map<Vector2i, double, CompareVec> f_score;

    open_set.insert(from);
    g_score[from] = 0;
    f_score[from] = 0;

    bool route_good = false;
    Vector2i best_goal(0, 0);
    double best_distance = 1e10;

    while (!open_set.empty()) {
        double current_best = 1e10;
        Vector2i current;
        for (const Vector2i &v: open_set) {
            if (f_score[v] < current_best) {
                current_best = f_score[v];
                current = v;
            }
        }

        Vector2d current_d = gm.convert_coord(current);
        if (goal.inside(current_d) && (current_d - center).norm() < best_distance) {
            best_goal = current;
            best_distance = (current_d - center).norm();
            route_good = true;
        }

        open_set.erase(current);

        // generate neighbors
        for (int i = -1; i <= 1; i++) {
            for (int j = -1; j <= 1; j++) {
                if (i == 0 && j == 0)
                    continue;


                Vector2i neighbor(current[0] + i, current[1] + j);
                if (!valid_cell(gm, neighbor))
                        continue;

                // g_score + d(current, neighbor);
                double tentive_g_score = g_score[current] + (std::abs(i) + std::abs(j));
                if (tentive_g_score < get_default(g_score, neighbor, 1e10)) {
                    // use distance to rect as a heuristic
                    double h = dist_to_rect(goal, gm.convert_coord(neighbor));

                    came_from[neighbor] = current;
                    g_score[neighbor] = tentive_g_score;
                    f_score[neighbor] = g_score[neighbor] + h;
                    open_set.insert(neighbor);
                }
            }
        }
    }

    // reconstruct path
    if (!route_good) {
        return false;
    }

    Vector2i current = best_goal;
    path.push_front(current);
    while (came_from.find(current) != came_from.end()) {
        current = came_from[current];
        path.push_front(current);
    }

    return true;
}

}

bool plan_route(std::list<Eigen::Vector2d> &route, const GridMap &gm,
        const Eigen::Vector2d &from, const Rect &dest)
{
    Vector2i from_valid = gm.convert_coord(from);
    find_valid_cell(gm, from_valid);

    std::list<Vector2i> path;
    if (!a_star(path, gm, from_valid, dest)) {
        warn("no route found");
        return false;
    }

    route.clear();

    // prune waypoints
    std::vector<Vector2i> path_vec(path.begin(), path.end());
    size_t begin = 0;
    while (begin < path_vec.size()) {
        size_t l = begin;
        size_t r = path_vec.size() - 1;
        while (l <= r) {
            size_t m = (l + r) / 2;

            // check of the vector between l and m is valid
            Vector2i x = path_vec[begin];
            Vector2i y = path_vec[m];
            Vector2i d = y - x;
            size_t steps = std::max(std::abs(d[0]), std::abs(d[1]));
            bool valid_route = true;
            for (size_t step = 0; step < steps; step++) {
                Vector2i v = x + (d.cast<double>() * step / steps).cast<int>();

                // this is always valid if x and y are valid
                assert(gm.valid_coord(v));

                if (!valid_cell(gm, v)) {
                    valid_route = false;
                    break;
                }
            }

            if (valid_route)
                l = m + 1;
            else
                r = m - 1;
        }

        // add the waypoint
        route.push_back(gm.convert_coord(path_vec[r]));
        begin = r + 1;
    }

    return true;
}
