#include "streamer.h"

#include <cassert>
#include <cerrno>
#include <cstdint>
#include <cstdlib>
#include <cstring>
#include <fcntl.h>
#include <iostream>
#include <csignal>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

namespace {

AVPixelFormat get_format_yuv420p(AVCodecContext *ctx, const enum AVPixelFormat *pix_fmts) {
    (void)ctx;

    AVPixelFormat fmt;
    while ((fmt = *(pix_fmts++)) != -1) {
        if (fmt == AV_PIX_FMT_YUV420P)
            return fmt;
    }

    return AV_PIX_FMT_NONE;
}

}

Streamer::Streamer() {
    m_parser_ctx = nullptr;
    m_decoder_ctx = nullptr;
    m_frame0 = nullptr;
    m_frame1 = nullptr;
    m_packet = nullptr;

    m_new_frame = false;

    m_width = -1;
    m_height = -1;

    memset(m_buffer, 0, STREAMER_BUFFER_SIZE + AV_INPUT_BUFFER_PADDING_SIZE);

    avcodec_register_all();

    const AVCodec *codec = avcodec_find_decoder(AV_CODEC_ID_H264);
    if (codec == nullptr) {
        clean_up();
        throw StreamerException("H264 codec not found");
    }

    m_parser_ctx = av_parser_init(codec->id);
    if (m_parser_ctx == nullptr) {
        clean_up();
        throw StreamerException("Failed to init parser context");
    }

    m_decoder_ctx = avcodec_alloc_context3(codec);
    if (m_decoder_ctx == nullptr) {
        clean_up();
        throw StreamerException("Failed to alloc decoder context");
    }
    m_decoder_ctx->get_format = get_format_yuv420p;

    if (avcodec_open2(m_decoder_ctx, codec, nullptr) < 0) {
        clean_up();
        throw StreamerException("Failed to open decoder");
    }

    m_packet = av_packet_alloc();
    if (m_packet == nullptr) {
        clean_up();
        throw StreamerException("Failed to alloc packet");
    }

    m_frame0 = av_frame_alloc();
    if (m_frame0 == nullptr) {
        clean_up();
        throw StreamerException("Failed to alloc frame");
    }

    m_frame1 = av_frame_alloc();
    if (m_frame1 == nullptr) {
        clean_up();
        throw StreamerException("Failed to alloc frame");
    }
}

Streamer::~Streamer() {
    clean_up();
}

void Streamer::push_data(const void *buf, size_t size) {
    if (size > STREAMER_BUFFER_SIZE)
        throw StreamerException("push_data(): size is too large");

    // we need the padding at the end of the buffer
    memcpy(m_buffer, buf, size);

    uint8_t *data = m_buffer;
    size_t data_size = size;
    while (data_size > 0) {
        int ret = av_parser_parse2(m_parser_ctx, m_decoder_ctx,
                &m_packet->data, &m_packet->size, data, data_size,
                AV_NOPTS_VALUE, AV_NOPTS_VALUE, 0);

        // this should not happen, i guess
        if (ret < 0)
            throw StreamerException("Error while parsing");

        data += ret;
        data_size -= ret;

        if (m_packet->size) {
            ret = avcodec_send_packet(m_decoder_ctx, m_packet);
            if (ret < 0)
                throw StreamerException("avcodec_send_packet(): Failed");

            while (ret >= 0) {
                ret = avcodec_receive_frame(m_decoder_ctx, m_frame1);
                if (ret == AVERROR(EAGAIN) || ret == AVERROR_EOF)
                    break;
                else if (ret < 0)
                    throw StreamerException("avcodec_receive_frame(): Failed");

                assert(ret == 0);
                assert(m_frame1->format == AV_PIX_FMT_YUV420P);

                AVFrame *tmp = m_frame0;
                m_frame0 = m_frame1;
                m_frame1 = tmp;

                m_new_frame = true;

                m_width = m_frame0->width;
                m_height = m_frame0->height;
            }
        }
    }
}

bool Streamer::get_frame(StreamerFrame &frame) {
    if (! m_new_frame)
        return false;
    m_new_frame = false;

    frame.width = m_frame0->width;
    frame.height = m_frame0->height;
    frame.y_plane = m_frame0->data[0];
    frame.u_plane = m_frame0->data[1];
    frame.v_plane = m_frame0->data[2];
    frame.y_pitch = m_frame0->linesize[0];
    frame.u_pitch = m_frame0->linesize[1];
    frame.v_pitch = m_frame0->linesize[2];

    return true;
}

void Streamer::get_size(int &width, int &height) {
    width = m_width;
    height = m_height;
}

void Streamer::clean_up() {
    av_packet_free(&m_packet);
    av_frame_free(&m_frame1);
    av_frame_free(&m_frame0);
    avcodec_free_context(&m_decoder_ctx);
    av_parser_close(m_parser_ctx);
}
