#include "streamer_widget.h"

#include "logging/logging.h"

#include <cassert>

StreamerWidget::StreamerWidget() :
    TextureComponent(SDL_PIXELFORMAT_IYUV, SDL_TEXTUREACCESS_STREAMING)
{
    m_streamer = nullptr;
}

void StreamerWidget::set_streamer(Streamer *streamer) {
    m_streamer = streamer;
}

Streamer *StreamerWidget::get_streamer() const {
    return m_streamer;
}

void StreamerWidget::request_size(WidgetSize &size, const WidgetSize &hint) {
    if (!m_streamer) {
        // if threre is no streamer, we just make the screen black and do not
        // request any size changes
        size = hint;
        return;
    }

    int stream_width;
    int stream_height;
    m_streamer->get_size(stream_width, stream_height);
    if (stream_width == -1 || stream_height == -1) {
        size = hint;
        return;
    }

    int new_width = hint.height * stream_width / stream_height;
    int new_height = hint.width * stream_height / stream_width;
    if (new_width <= hint.width) {
        size.width = new_width;
        size.height = hint.height;
    }
    else {
        size.width = hint.width;
        size.height = new_height;
    }
}

void StreamerWidget::draw(SDL_Renderer *renderer, const WidgetDrawLocation &location) {
    SDL_Rect clear_rect = {location.x, location.y, location.size.width, location.size.height};
    SDL_SetRenderDrawColor(renderer, 0x00, 0x00, 0x00, 0xff);
    SDL_RenderFillRect(renderer, &clear_rect);

    if (!m_streamer)
        return;

    StreamerFrame frame;
    bool new_frame = m_streamer->get_frame(frame);

    if (new_frame)
        update_texture(renderer, frame);

    if (!m_texture)
        return;

    int new_width = location.size.height * m_texture_width / m_texture_height;
    int new_height = location.size.width * m_texture_height / m_texture_width;

    SDL_Rect dstrect;
    if (new_width <= location.size.width) {
        dstrect.w = new_width;
        dstrect.h = location.size.height;
    }
    else {
        dstrect.w = location.size.width;
        dstrect.h = new_height;
    }

    int horizontal_space = location.size.width - dstrect.w;
    int vertical_space = location.size.height - dstrect.h;
    assert(horizontal_space >= 0);
    assert(vertical_space >= 0);
    dstrect.x = location.x + horizontal_space / 2;
    dstrect.y = location.y + vertical_space / 2;

    SDL_RenderCopy(renderer, m_texture, nullptr, &dstrect);
}

void StreamerWidget::update_texture(SDL_Renderer *renderer, const StreamerFrame &frame) {
    allocate_texture(renderer, frame.width, frame.height);
    if (m_texture == nullptr)
        return;

    int ret = SDL_UpdateYUVTexture(m_texture, nullptr,
            frame.y_plane, frame.y_pitch,
            frame.u_plane, frame.u_pitch,
            frame.v_plane, frame.v_pitch);
    if (ret != 0)
        error("SDL_UpdateYUVTexture()", SDL_GetError());
}

