#include "network.h"

#include <stdio.h>
#include <iostream>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <cerrno>
#include <unistd.h>
#include <cstring>
#include <sys/socket.h>
#include <system_error>
#include <arpa/inet.h>


namespace {

class EndOfStreamException : public std::exception {
    public:
        EndOfStreamException() {}
};

size_t my_send(int fd, const void *buf, size_t size) {
    ssize_t ret = send(fd, buf, size, 0);

    if (ret == -1) {
        if (errno == EPIPE)
            throw EndOfStreamException();
        throw std::system_error(errno, std::generic_category());
    }
    else if (ret == 0)
        throw EndOfStreamException();

    return ret;
}

void safe_send(int fd, const void *buf, size_t size) {
    const char *p = (const char *)buf;
    const char *end = p + size;

    while (p < end) {
        ssize_t ret = send(fd, p, end - p, 0);
        if (ret == -1) {
            if (errno == EPIPE)
                throw EndOfStreamException();
            throw std::system_error(errno, std::generic_category());
        }
        else if (ret == 0)
            throw EndOfStreamException();

        p += ret;
    }
}

void safe_recv(int fd, void *buf, size_t size) {
    char *p = (char *)buf;
    char *end = p + size;

    while (p < end) {
        ssize_t ret = recv(fd, p, end - p, 0);
        if (ret == -1) {
            if (errno == EPIPE)
                throw EndOfStreamException();
            throw std::system_error(errno, std::generic_category());
        }
        else if (ret == 0)
            throw EndOfStreamException();

        p += ret;
    }
}

}


Network::Network(const char *ip_addr, short port, bool server) {
    m_fd = -1;
    m_connected = true;
    m_buffer = nullptr;

    int sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock == -1)
        throw std::system_error(errno, std::generic_category());

    struct sockaddr_in addr;
    memset(&addr, 0, sizeof(addr));

    addr.sin_family = AF_INET;
    inet_pton(AF_INET, ip_addr, &(addr.sin_addr));
    addr.sin_port = htons(port);

    if (server) {
        if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) == -1) {
            close(sock);
            throw std::system_error(errno, std::generic_category());
        }

        if (listen(sock, 1) == -1) {
            close(sock);
            throw std::system_error(errno, std::generic_category());
        }

        m_fd = accept(sock, NULL, NULL);
        if (m_fd == -1) {
            close(sock);
            throw std::system_error(errno, std::generic_category());
        }

        std::cout << "Accepted a new connection" << std::endl;

        close(sock);
    }
    else {
        if (connect(sock, (struct sockaddr *)&addr, sizeof(addr)) == -1)
        {
            close(sock);
            throw std::system_error(errno, std::generic_category());
        }

        m_fd = sock;
    }

    // allocate send buffer
    m_buffer = new uint8_t[NETWORK_BUFFER_MAX_SIZE];
    m_buffer_start = 0;
    m_buffer_end = 0;
    m_buffer_size = 0;
}

Network::~Network() {
    if (m_connected) {
        try {
            buffer_send_all();
        }
        catch (EndOfStreamException &e) {
            m_connected = false;
        }
    }

    if (m_fd != -1)
        close(m_fd);

    delete[] m_buffer;
}


void Network::send_pkt(uint8_t type, const void *buf, size_t size) {
    if (size > NETWORK_MAX_PAYLOAD_SIZE)
        throw std::invalid_argument("Payload to large");

    if (! m_connected)
        return;

    try {
        buffer_push_back(&type, sizeof(type));

        uint16_t payload_size = htons(size);
        buffer_push_back(&payload_size, sizeof(payload_size));

        buffer_push_back(buf, size);
    }
    catch(EndOfStreamException &e) {
        m_connected = false;
    }

    buffer_send();
}

bool Network::recv_pkt(uint8_t &type, void *buf, size_t &size) {
    if (! m_connected)
        return false;

    // check if there is data available
    fd_set recv_set;
    FD_ZERO(&recv_set);
    FD_SET(m_fd, &recv_set);

    struct timeval tv = {0, 0}; // return immediately

    int ret = select(m_fd + 1, &recv_set, NULL, NULL, &tv);
    if (ret == -1)
        throw std::system_error(errno, std::generic_category());

    if (! FD_ISSET(m_fd, &recv_set))
        return false;

    // recv packet
    try {
        safe_recv(m_fd, &type, sizeof(type));

        uint16_t payload_size = 0;
        safe_recv(m_fd, &payload_size, sizeof(payload_size));
        payload_size = ntohs(payload_size);

        if (payload_size > size)
            throw std::invalid_argument("Buffer not large enough");
        size = payload_size;

        safe_recv(m_fd, buf, size);
    }
    catch (EndOfStreamException &e) {
        m_connected = false;
    }

    return m_connected;
}


bool Network::connected() {
    return m_connected;
}

void Network::buffer_push_back(const void *data, size_t size) {
    size_t size_avail = NETWORK_BUFFER_MAX_SIZE - m_buffer_size;
    if (size >= size_avail)
        throw std::overflow_error("send buffer overflow");

    const uint8_t *ptr = (const uint8_t *)data;
    m_buffer_size += size;

    // copy until end of m_buffer
    size_t size_cpy = std::min(NETWORK_BUFFER_MAX_SIZE - m_buffer_end, size);
    memcpy(&(m_buffer[m_buffer_end]), ptr, size_cpy);

    m_buffer_end = (m_buffer_end + size_cpy) % NETWORK_BUFFER_MAX_SIZE;
    ptr += size_cpy;
    size -= size_cpy;

    // copy from the start of the buffer (if any data is still not copied)
    memcpy(&(m_buffer[m_buffer_end]), ptr, size);
    m_buffer_end = (m_buffer_end + size) % NETWORK_BUFFER_MAX_SIZE;
}

void Network::buffer_send() {
    if (!m_connected)
        return;

    fd_set send_set;
    FD_ZERO(&send_set);
    FD_SET(m_fd, &send_set);

    struct timeval tv = {0, 0};

    int ret = select(m_fd + 1, NULL, &send_set, NULL, &tv);
    if (ret == -1)
        throw std::system_error(errno, std::generic_category());

    if (!FD_ISSET(m_fd, &send_set))
        return;

    size_t size = std::min(NETWORK_BUFFER_MAX_SIZE - m_buffer_start, m_buffer_size);

    try {
        size_t bytes_sent = my_send(m_fd, &(m_buffer[m_buffer_start]), size);
        m_buffer_size -= bytes_sent;
        m_buffer_start = (m_buffer_start + size) % NETWORK_BUFFER_MAX_SIZE;
    }
    catch (EndOfStreamException &e) {
        m_connected = false;
    }
}

void Network::buffer_send_all() {
    if (!m_connected)
        return;

    // send everything until the end of the buffer
    size_t size = std::min(NETWORK_BUFFER_MAX_SIZE - m_buffer_start, m_buffer_size);
    try {
        safe_send(m_fd, &(m_buffer[m_buffer_start]), size);
        m_buffer_start = (m_buffer_start + size) % NETWORK_BUFFER_MAX_SIZE;
        m_buffer_size -= size;
    }
    catch(EndOfStreamException &e) {
        m_connected = false;
        return;
    }

    // send data at the start of the buffer
    try {
        safe_send(m_fd, &(m_buffer[m_buffer_start]), m_buffer_size);
        m_buffer_size = 0;
    }
    catch (EndOfStreamException &e) {
        m_connected = false;
        return;
    }
}

