// Ceres Solver - A fast non-linear least squares minimizer
// Copyright 2016 Google Inc. All rights reserved.
// http://ceres-solver.org/
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// * Redistributions of source code must retain the above copyright notice,
//   this list of conditions and the following disclaimer.
// * Redistributions in binary form must reproduce the above copyright notice,
//   this list of conditions and the following disclaimer in the documentation
//   and/or other materials provided with the distribution.
// * Neither the name of Google Inc. nor the names of its contributors may be
//   used to endorse or promote products derived from this software without
//   specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// Author: vitus@google.com (Michael Vitus)

#include <Eigen/Core>

#include <ceres/ceres.h>

#include "ceres_solver.h"

using namespace Eigen;

namespace {

template <typename T> Matrix<T, 2, 2> rotation_matrix(T yaw) {
    const T cos_yaw = ceres::cos(yaw);
    const T sin_yaw = ceres::sin(yaw);

    Matrix<T, 2, 2> rotation;
    rotation << cos_yaw, -sin_yaw, sin_yaw, cos_yaw;
    return rotation;
}

template <typename T> inline T normalize_angle(const T& angle) {
  T two_pi(2.0 * M_PI);
  T wholes = ceres::floor((angle + T(M_PI)) / two_pi);
  return angle - two_pi * wholes;
}

class AngleLocalParameterization {
    public:
        template <typename T>
        bool operator()(const T* theta, const T* delta_theta, T* theta_plus_delta) const {
            *theta_plus_delta = normalize_angle(*theta + *delta_theta);

            return true;
        }

        static ceres::LocalParameterization* create() {
            return (new ceres::AutoDiffLocalParameterization<AngleLocalParameterization, 1, 1>);
        }
};

class PoseGraph2dErrorTerm {
    public:
        PoseGraph2dErrorTerm(double x_ab, double y_ab, double yaw_ab, const Matrix3d& information) :
            p_ab_(x_ab, y_ab), yaw_ab_(yaw_ab), information_(information) {}

        template <typename T>
        bool operator()(const T* const x_a, const T* const y_a, const T* const yaw_a,
          const T* const x_b, const T* const y_b, const T* const yaw_b, T* residuals_ptr) const
        {
            const Matrix<T, 2, 1> p_a(*x_a, *y_a);
            const Matrix<T, 2, 1> p_b(*x_b, *y_b);

            Map<Eigen::Matrix<T, 3, 1>> residuals_map(residuals_ptr);

            residuals_map.template head<2>() =
                rotation_matrix(*yaw_a).transpose() * (p_b - p_a) - p_ab_.cast<T>();
            residuals_map(2) = normalize_angle((*yaw_b - *yaw_a) - static_cast<T>(yaw_ab_));

            residuals_map = information_.template cast<T>() * residuals_map;

            return true;
        }

        static ceres::CostFunction* create(double x_ab, double y_ab,
                double yaw_ab, const Eigen::Matrix3d& information)
        {
            return (new ceres::AutoDiffCostFunction<PoseGraph2dErrorTerm, 3, 1, 1, 1, 1, 1, 1>(
                        new PoseGraph2dErrorTerm(x_ab, y_ab, yaw_ab, information)));
        }

        EIGEN_MAKE_ALIGNED_OPERATOR_NEW

    private:
        const Vector2d p_ab_;
        const double yaw_ab_;
        const Matrix3d information_;
};

void build_optimization_problem(const std::vector<Constraint2d>& constraints,
          std::map<int, Pose2d>* poses, ceres::Problem* problem) {
    assert(poses != NULL);
    assert(problem != NULL);
    assert(!constraints.empty());

    ceres::LossFunction* loss_function = NULL;
    ceres::LocalParameterization* angle_local_parameterization = AngleLocalParameterization::create();

    for (const Constraint2d &constraint: constraints) {
        std::map<int, Pose2d>::iterator pose_begin_iter = poses->find(constraint.id_begin);
        std::map<int, Pose2d>::iterator pose_end_iter = poses->find(constraint.id_end);
        assert(pose_begin_iter != poses->end());
        assert(pose_end_iter != poses->end());

        const Eigen::Matrix3d sqrt_information = constraint.information.llt().matrixL();
        // Ceres will take ownership of the pointer.
        ceres::CostFunction* cost_function = PoseGraph2dErrorTerm::create(
            constraint.x, constraint.y, constraint.yaw_radians, sqrt_information);
        problem->AddResidualBlock(cost_function, loss_function,
                                  &pose_begin_iter->second.x,
                                  &pose_begin_iter->second.y,
                                  &pose_begin_iter->second.yaw_radians,
                                  &pose_end_iter->second.x,
                                  &pose_end_iter->second.y,
                                  &pose_end_iter->second.yaw_radians);

        problem->SetParameterization(&pose_begin_iter->second.yaw_radians,
                angle_local_parameterization);
        problem->SetParameterization(&pose_end_iter->second.yaw_radians,
                angle_local_parameterization);
    }

    std::map<int, Pose2d>::iterator pose_start_iter = poses->begin();
    assert(pose_start_iter != poses->end());
    problem->SetParameterBlockConstant(&pose_start_iter->second.x);
    problem->SetParameterBlockConstant(&pose_start_iter->second.y);
    problem->SetParameterBlockConstant(&pose_start_iter->second.yaw_radians);
}

// Returns true if the solve was successful.
bool solve_optimization_problem(ceres::Problem* problem) {
    assert(problem != NULL);

    ceres::Solver::Options options;
    options.max_num_iterations = 100;
    options.linear_solver_type = ceres::SPARSE_NORMAL_CHOLESKY;

    ceres::Solver::Summary summary;
    ceres::Solve(options, problem, &summary);

    return summary.IsSolutionUsable();
}

}

bool optimize_pose_graph(std::map<int, Pose2d> &poses, std::vector<Constraint2d> &constraints) {
    ceres::Problem problem;
    build_optimization_problem(constraints, &poses, &problem);

    return solve_optimization_problem(&problem);
}
