#include "auto_free.h"
#include "graph.h"
#include "passes.h"
#include "misc.h"
#include "ceres_solver.h"

#include "logging/logging.h"
#include "logging/except.h"

#include <vector>
#include <map>
#include <set>
#include <limits>

using Eigen::Vector2i;
using Eigen::Vector2d;
using Eigen::Vector3d;
using Eigen::Matrix3d;
using Eigen::Array3d;
using Eigen::Array2i;

namespace {

const size_t ROBOT_HISTORY_UNKOWN = std::numeric_limits<size_t>::max();

const double FLOODFILL_FACTOR    = 1.0;
const double FLOODFILL_THRESHOLD = 0.5;

void gm_draw_bounds(GridMap &gm, const Rect &bounds, Matrix3d transform = Matrix3d::Identity()) {
    double xmin = bounds.pos[0];
    double xmax = xmin + bounds.dim[0];
    double ymin = bounds.pos[1];
    double ymax = ymin + bounds.dim[1];

    Vector2d verts[4];
    verts[0] = Vector2d(xmin, ymin);
    verts[1] = Vector2d(xmin, ymax);
    verts[2] = Vector2d(xmax, ymax);
    verts[3] = Vector2d(xmax, ymin);

    for (size_t i = 0; i < 4; i++) {
        Vector3d v(verts[i][0], verts[i][1], 1.0);
        verts[i] = (transform * v).head(2);
    }

    for (size_t i = 0; i < 4; i++)
        gm.draw_line(verts[i], verts[(i + 1) % 4]);
}

void lm_export(const LocalMap &lm) {
    char fname[128];
    snprintf(fname, sizeof(fname), "./%d_%d.gm", lm.get_start_index(), lm.get_end_index());
    lm.get_gm().write(fname);
}

}

Graph::Graph() :
    m_robot_history(50.0 / 100.0, 128),
    m_ggm(2.0 / 100, 1024),
    m_ggm_prev(2.0 / 100, 1024)
{
    m_draw_bounds       = true;
    m_export_local_maps = false;
    m_draw_robot_trail  = true;
    m_count = 0;
    m_ggm_dirty = false;
    m_ggm_requires_floodfill = false;

    m_robot_history.fill(ROBOT_HISTORY_UNKOWN);
}

void Graph::write(const std::string &filename) const {
    AutoFree<FILE *, int> file(fopen(filename.c_str(), "w"), fclose);
    if (file == nullptr)
        throw FM_SYS("fopen");

    for (size_t i = 0; i < m_local_maps.size(); i++) {
        const Matrix3d &pose = m_local_maps[i].get_global_pose();
        int ret = fprintf(file, "VERTEX_SE2 %d %lf %lf %lf\n", (int)i,
                pose(0, 2), pose(1, 2), std::atan2(pose(1, 0), pose(0, 0)));
        if (ret < 0)
            throw FM_SYS("fprintf");
    }

    for (size_t i = 0; i < m_constraints.size(); i++) {
        const Constraint &c = m_constraints[i];
        int ret = fprintf(file, "EDGE_SE2 %d %d %lf %lf %lf 1.0 0.0 0.0 1.0 0.0 1.0\n",
                (int)c.i, (int)c.j, c.t(0, 2), c.t(1, 2), std::atan2(c.t(1, 0), c.t(0, 0)));
        if (ret < 0)
            throw FM_SYS("fprintf");
    }
}

void Graph::insert_pc(const PointCloud &pc) {
    if (m_count % 50 == 0) {
        // finalize the current map
        finalize();

        // create a new map
        Matrix3d pose = Matrix3d::Identity();
        if (m_count != 0) {
            LocalMap &map = m_local_maps.back();
            pose = map.get_current_pose() * map.get_global_pose();
        }

        LocalMap new_map(m_count, m_count + 50, pose, 2.0 / 100, 1024);
        m_local_maps.push_back(new_map);
        m_ggm_lm_positions[m_local_maps.size() - 1] = mul_m3_v2(pose, Vector2d::Zero());
    }

    LocalMap &map = m_local_maps.back();
    map.insert_pc(pc);

    Matrix3d robot_pose = map.get_global_pose() * map.get_current_pose();
    Vector2d robot_location(robot_pose(0, 2), robot_pose(1, 2));
    m_robot_history[robot_location] = m_local_maps.size() - 1;

    PointCloud global_pc = map.get_point_clouds().back();
    global_pc.transform(map.get_global_pose());
    pass_gm_insert_pc(m_ggm, global_pc);
    m_ggm.push_back_robot(Vector2d(robot_pose(0, 2), robot_pose(1, 2)));

    m_count += 1;
}

void Graph::get_gm(GridMap &gm) {
    assert(m_local_maps.size() > 0);

    if (m_ggm_dirty)
        ggm_patch();

    if (m_ggm_requires_floodfill) {
        pass_gm_floodfill_clean(m_ggm, FLOODFILL_FACTOR, FLOODFILL_THRESHOLD);
        m_ggm_requires_floodfill = false;
    }

    gm = m_ggm;
    if (m_draw_bounds) {
        const LocalMap &map = m_local_maps.back();
        gm_draw_bounds(gm, map.get_gm().get_bounding_box(), map.get_global_pose());
    }

    // draw the most recent robot positions
    gm.clear_robot_positions();
    if (m_draw_robot_trail) {
        size_t start = 0;
        size_t end = m_local_maps.size();
        if (end > 10)
            start = end - 10;

        for (size_t i = start; i < end; i++) {
            const LocalMap &lm = m_local_maps[i];
            Matrix3d pose = lm.get_global_pose();

            std::vector<Vector2d> robots;
            lm.get_gm().get_robot_positions(robots);

            for (Vector2d robot: robots)
                gm.push_back_robot(mul_m3_v2(pose, robot));
        }
    }

    // draw the most recent annotation map
    int width2 = m_robot_history.get_width() / 2;
    int height2 = m_robot_history.get_height() / 2;
    for (int y = -height2; y < height2; y++) {
        for (int x = -width2; x < width2; x++) {
            Vector2i c(x + width2 - gm.get_width() / 2, y + height2 - gm.get_height() / 2);
            if (m_robot_history[Vector2i(x, y)] != ROBOT_HISTORY_UNKOWN)
                gm.annotation(c) = 1.0;
            else
                gm.annotation(c) = 0.5;
        }
    }
}

void Graph::get_robot_position(Eigen::Vector2d &vec) {
    if (m_local_maps.size() == 0) {
        vec = Vector2d(0.0, 0.0);
        return;
    }

    LocalMap &map = m_local_maps.back();
    Matrix3d global_pose = map.get_global_pose();
    Matrix3d local_pose = map.get_current_pose();
    Matrix3d pose = local_pose * global_pose;

    vec = mul_m3_v2(pose, Vector2d(0.0, 0.0));
}

void Graph::finalize() {
    if (m_local_maps.size() < 1)
        return;

    LocalMap &map = m_local_maps.back();

    pass_gm_floodfill_clean(map.get_gm(), FLOODFILL_FACTOR, FLOODFILL_THRESHOLD);

    if (m_draw_bounds)
        gm_draw_bounds(map.get_gm(), map.get_gm().get_bounding_box());

    if (m_export_local_maps)
        lm_export(map);

    m_ggm_requires_floodfill = true;

    loopclosure();
    optimize();
}

void Graph::loopclosure() {
    assert(m_local_maps.size() > 0);

    // alias the map we want to loopclose
    LocalMap &map = m_local_maps.back();

    // find collisions based om bounding box
    std::vector<size_t> collissions;
    Rect map_bounds = map.get_gm().get_bounding_box();
    for (size_t i = 0; i < m_local_maps.size() - 1; i++) {
        Rect bounds = m_local_maps[i].get_gm().get_bounding_box();

        if (map_bounds.pos[0] + map_bounds.dim[0] >= bounds.pos[0]                 &&
            map_bounds.pos[0]                     <= bounds.pos[0] + bounds.dim[0] &&
            map_bounds.pos[1] + map_bounds.dim[1] >= bounds.pos[1]                 &&
            map_bounds.pos[1]                     <= bounds.pos[1] + bounds.dim[1])
        {
            collissions.push_back(i);
        }
    }

    PointCloud pc;
    map.get_gm().get_point_cloud(pc);
    pc.transform(map.get_global_pose());
    for (size_t i: collissions) {
        Matrix3d pose = m_local_maps[i].get_global_pose().inverse();
        double error = align_pc_to_gm2(&(m_local_maps[i].get_gm()), &pc, pose);
        pose *= map.get_global_pose();
        if (error < 0.9)
            m_constraints.push_back({i, m_local_maps.size() - 1, pose, error});
    }

    // constraint obtained from consecutive gridmaps
    if (m_local_maps.size() > 1) {
        size_t i = m_local_maps.size() - 2;
        size_t j = m_local_maps.size() - 1;
        Matrix3d pose = m_local_maps[i].get_global_pose().inverse() *
            m_local_maps[j].get_global_pose();
        m_constraints.push_back({i, j, pose, 1.0});
    }
}

void Graph::optimize() {
    if (m_local_maps.size() < 2)
        return;

    info("poses:", m_local_maps.size());
    info("constraints:", m_constraints.size());

    std::map<int, Pose2d> poses;
    std::vector<Constraint2d> constraints;

    for (size_t i = 0; i < m_local_maps.size(); i++) {
        const Matrix3d &p = m_local_maps[i].get_global_pose();

        Pose2d pose;
        pose.x = p(0, 2);
        pose.y = p(1, 2);
        pose.yaw_radians = std::atan2(p(1, 0), p(0, 0));

        poses[i] = pose;
    }

    for (size_t i = 0; i < m_constraints.size(); i++) {
        const Constraint &c = m_constraints[i];

        Constraint2d cons;
        cons.id_begin = c.i;
        cons.id_end = c.j;
        cons.x = c.t(0, 2);
        cons.y = c.t(1, 2);
        cons.yaw_radians = std::atan2(c.t(1, 0), c.t(0, 0));
        cons.information = Matrix3d::Identity();

        constraints.push_back(cons);
    }

    optimize_pose_graph(poses, constraints);

    bool restore_from_prev = false;
    for (size_t i = 0; i < poses.size(); i++) {
        double x = poses[i].x;
        double y = poses[i].y;
        double th = poses[i].yaw_radians;

        Matrix3d p = Matrix3d::Identity();
        p(0, 0) = std::cos(th);
        p(1, 1) = std::cos(th);
        p(0, 1) = -std::sin(th);
        p(1, 0) = std::sin(th);
        p(0, 2) = x;
        p(1, 2) = y;

        m_local_maps[i].set_global_pose(p);

        // check if there are maps drawn at the wrong location in the
        // m_ggm map
        Vector2d graph_location = mul_m3_v2(p, Vector2d::Zero());
        Vector2d ggm_location = m_ggm_lm_positions[i];
        Vector2d diff = graph_location - ggm_location;
        Array2i diff_abs = static_cast<Vector2i>(m_ggm.convert_coord(diff)).array().abs();
        if (diff_abs[0] > 0 || diff_abs[1] > 0) {
            info("position in ggm changed:", i, m_local_maps.size());
            if (i == m_local_maps.size() - 1)
                restore_from_prev = true;
            else
                update_ggm_patch(m_local_maps[i].get_gm().get_bounding_box());
        }
    }

    if (restore_from_prev) {
        if (m_ggm_dirty) {
            update_ggm_patch(m_local_maps.back().get_gm().get_bounding_box());
        }
        else {
            restore_prev();
        }
    }

    if (!m_ggm_dirty)
        m_ggm_prev = m_ggm;

    // update the history map
    m_robot_history.fill(ROBOT_HISTORY_UNKOWN);
    for (size_t i = 0; i < m_local_maps.size(); i++) {
        const LocalMap &lm = m_local_maps[i];

        std::vector<Vector2d> robot_positions;
        lm.get_gm().get_robot_positions(robot_positions);

        Matrix3d pose = lm.get_global_pose();
        for (Vector2d robot: robot_positions)
            m_robot_history[mul_m3_v2(pose, robot)] = i;
    }
}

void Graph::update_ggm_patch(const Rect &patch) {
    if (!m_ggm_dirty) {
        m_ggm_patch = patch;
        m_ggm_dirty = true;
        return;
    }

    double xmin = std::min(m_ggm_patch.get_xmin(), patch.get_xmin());
    double xmax = std::max(m_ggm_patch.get_xmax(), patch.get_xmax());
    double ymin = std::min(m_ggm_patch.get_ymin(), patch.get_ymin());
    double ymax = std::max(m_ggm_patch.get_ymax(), patch.get_ymax());

    m_ggm_patch = Rect(xmin, xmax, ymin, ymax);
}

void Graph::ggm_patch() {
    assert(m_ggm_dirty);

    m_ggm.clear(m_ggm_patch);
    m_ggm.clear_annotations();

    const int width2 = m_robot_history.get_width() / 2;
    const int height2 = m_robot_history.get_height() / 2;

    // get the currently active maps
    std::set<size_t> active_maps;
    for (int y = -height2; y < height2; y++) {
        for (int x = -width2; x < width2; x++) {
            Vector2i v(x, y);
            if (m_robot_history[v] != ROBOT_HISTORY_UNKOWN)
                active_maps.insert(m_robot_history[v]);
        }
    }

    for (size_t i: active_maps) {
        Rect bounds = m_local_maps[i].get_gm().get_bounding_box();
        bool redraw_map = m_ggm_patch.overlaps(bounds);

        const LocalMap &map = m_local_maps[i];
        for (PointCloud pc: map.get_point_clouds()) {
            pc.transform(map.get_global_pose());

            if (redraw_map) {
                pass_gm_insert_pc(m_ggm, pc, m_ggm_patch);
            }

            if (m_ggm_patch.inside(pc.get_center()))
                m_ggm.push_back_robot(pc.get_center());
        }

        if (redraw_map)
            m_ggm_lm_positions[i] = mul_m3_v2(map.get_global_pose(), Vector2d::Zero());
    }

    m_ggm_dirty = false;
    m_ggm_requires_floodfill = true;

    m_ggm_prev = m_ggm;
}

void Graph::restore_prev() {
    assert(! m_ggm_dirty);

    m_ggm = m_ggm_prev;

    const LocalMap &map = m_local_maps.back();
    for (PointCloud pc: map.get_point_clouds()) {
        pc.transform(map.get_global_pose());

        pass_gm_insert_pc(m_ggm, pc, m_ggm_patch);
        m_ggm.push_back_robot(pc.get_center());
    }

    m_ggm_lm_positions[m_local_maps.size() - 1] = mul_m3_v2(map.get_global_pose(), Vector2d::Zero());
}

