#include "grid_map.h"
#include "logging/logging.h"
#include "logging/except.h"

#include "lodepng/lodepng.h"

using namespace Eigen;

GridMap::GridMap(double cell_size, size_t grid_size) :
    m_cell_size(cell_size),
    m_grid_size(grid_size),
    m_hit_map(m_grid_size * m_grid_size, 0),
    m_miss_map(m_grid_size * m_grid_size, 0),
    m_annotation_map(m_grid_size * m_grid_size, 0)
{ }

void GridMap::clear() {
    std::fill(m_hit_map.begin(),        m_hit_map.end(),        0.0);
    std::fill(m_miss_map.begin(),       m_miss_map.end(),       0.0);
    std::fill(m_annotation_map.begin(), m_annotation_map.end(), 0.0);
    m_robot_positions.clear();
}

void GridMap::clear(const Rect &rect) {
    Vector2i min = convert_coord(rect.pos);
    Vector2i max = convert_coord(static_cast<Vector2d>(rect.pos + rect.dim));

    for (int y = min[1]; y < max[1]; y++) {
        for (int x = min[0]; x < max[0]; x++) {
            hit(Vector2i(x, y)) = 0.0;
            miss(Vector2i(x, y)) = 0.0;
        }
    }

    std::set<Vector2i, CompareVec> robot_positions;
    for (const Vector2i &r: m_robot_positions) {
        if (rect.inside(convert_coord(r)))
            continue;

        robot_positions.insert(r);
    }
    m_robot_positions = robot_positions;
}

void GridMap::clear_robot_positions() {
    m_robot_positions.clear();
}

void GridMap::clear_annotations() {
    std::fill(m_annotation_map.begin(), m_annotation_map.end(), 0.0);
}

void GridMap::read(const std::string &filename) {
    throw FM("Not implemented");
}

void GridMap::write(const std::string &filename) const {
    const char *ext = strrchr(filename.c_str(), '.');
    if (ext == nullptr)
        throw FM("No output format specified");


    if (strcmp(ext, ".gm") == 0)
        write_gm(filename);
    else if (strcmp(ext, ".png") == 0)
        write_png(filename);
    else
        throw FM("Output format not supported");
}

double GridMap::get_cell_size() const {
    return m_cell_size;
}

void GridMap::draw_line(const Eigen::Vector2i &v0, const Eigen::Vector2i &v1) {
    Array2i dv = v1 - v0;
    int steps = std::max(std::abs(dv[0]), std::abs(dv[1]));
    for (int i = 0; i < steps; i++) {
        Array2d dx = (i * dv.cast<double>() / steps).array().round();
        Vector2i x = v0 + dx.cast<int>().matrix();
        annotation(x) = 1.0;
    }
}

void GridMap::draw_line(const Eigen::Vector2d &v0, const Eigen::Vector2d &v1) {
    draw_line(convert_coord(v0), convert_coord(v1));
}

void GridMap::get_robot_positions(std::vector<Eigen::Vector2i> &positions) const {
    positions.clear();

    for (const Vector2i &robot: m_robot_positions) {
        if (! valid_coord(robot))
            continue;
        positions.push_back(robot);
    }
}

void GridMap::get_robot_positions(std::vector<Eigen::Vector2d> &positions) const {
    positions.clear();

    for (const Vector2i &robot: m_robot_positions) {
        if (! valid_coord(robot))
            continue;
        positions.push_back(convert_coord(robot));
    }
}

Rect GridMap::get_bounding_box(double threshold) const {
    int xmin = INT_MAX;
    int xmax = INT_MIN;
    int ymin = INT_MAX;
    int ymax = INT_MIN;

    bool good = false;

    int size2 = m_grid_size / 2;
    for (int y = -size2; y < size2; y++) {
        for (int x = -size2; x < size2; x++) {
            Vector2i coord(x, y);
            if (hit(coord) > threshold || miss(coord) > threshold) {
                xmin = std::min(x, xmin);
                xmax = std::max(x, xmax);
                ymin = std::min(y, ymin);
                ymax = std::max(y, ymax);

                good = true;
            }
        }
    }

    if (!good) {
        xmin = 0;
        xmax = 0;
        ymin = 0;
        ymax = 0;
    }

    Rect bounds;
    bounds.pos = convert_coord(Vector2i(xmin, ymin));
    bounds.dim = convert_coord(Vector2i(xmax - xmin, ymax - ymin));
    return bounds;
}

void GridMap::get_point_cloud(PointCloud &pc, double threshold) const {
    pc.clear();

    int size2 = m_grid_size / 2;
    for (int y = -size2; y < size2; y++) {
        for (int x = -size2; x < size2; x++) {
            if (hit(Vector2i(x, y)) > threshold) {
                PointCloudDatum datum;
                datum.timestamp = 0; // the timestamp is not well defined
                datum.data = convert_coord(Vector2i(x, y));

                pc.push_back(datum);
            }
        }
    }
}

void GridMap::write_gm(const std::string &filename) const {
    throw FM("Not Implemented");
}

void GridMap::write_png(const std::string &filename) const {
    std::vector<uint8_t> image(4 * m_grid_size * m_grid_size);

    for (size_t y = 0; y < m_grid_size; y++) {
        for (size_t x = 0; x < m_grid_size; x++) {
            uint8_t red     = 255 * hit(Vector2i(x - m_grid_size / 2, y - m_grid_size / 2));
            uint8_t green   = 255 * miss(Vector2i(x - m_grid_size / 2, y - m_grid_size / 2));
            uint8_t blue    = 0;

            double a = annotation(Vector2i(x - m_grid_size / 2, y - m_grid_size / 2));
            if (a > 0.1) {
                red     = 255 * a;
                green   = 255 * a;
                blue    = 255 * a;
            }

            image[4 * (x + y * m_grid_size) + 0] = red;
            image[4 * (x + y * m_grid_size) + 1] = green;
            image[4 * (x + y * m_grid_size) + 2] = blue;
            image[4 * (x + y * m_grid_size) + 3] = 0xff;
        }
    }

    for (const Vector2i &robot: m_robot_positions) {
        Vector2i vec = robot + Vector2i(m_grid_size / 2, m_grid_size / 2);
        image[4 * (vec(0) + vec(1) * m_grid_size) + 0] = 0x00;
        image[4 * (vec(0) + vec(1) * m_grid_size) + 1] = 0x00;
        image[4 * (vec(0) + vec(1) * m_grid_size) + 2] = 0xff;
        image[4 * (vec(0) + vec(1) * m_grid_size) + 3] = 0xff;
    }

    lodepng::encode(filename, image, m_grid_size, m_grid_size);
}

