#ifndef __grid_map_h__
#define __grid_map_h__

#include <Eigen/Dense>

#include <cstddef>
#include <cassert>
#include <cmath>
#include <vector>
#include <set>

#include "point_cloud.h"
#include "rect.h"
#include "misc.h"

class GridMap {
    public:
        GridMap(double cell_size, size_t grid_size);

        void clear();
        void clear(const Rect &rect);
        void clear_robot_positions();
        void clear_annotations();

        void read(const std::string &filename);
        void write(const std::string &filename) const;

        double get_cell_size() const;
        Eigen::Vector2i convert_coord(const Eigen::Vector2d &vec) const {
            Eigen::Vector2d x = vec / m_cell_size;
            return Eigen::Vector2i(std::round(x(0)), std::round(x(1)));
        }
        Eigen::Vector2d convert_coord(const Eigen::Vector2i &vec) const {
            Eigen::Vector2d x(vec(0), vec(1));
            return x * m_cell_size;
        }
        size_t get_width() const {
            return m_grid_size;
        }
        size_t get_height() const {
            return m_grid_size;
        }

        bool valid_coord(const Eigen::Vector2i &vec) const {
            size_t x = vec(0) + m_grid_size / 2;
            size_t y = vec(1) + m_grid_size / 2;
            return x < m_grid_size && y < m_grid_size;
        }
        bool valid_coord(const Eigen::Vector2d &vec) const {
            return valid_coord(convert_coord(vec));
        }

        const double &hit(const Eigen::Vector2i &vec) const {
            size_t x = vec(0) + m_grid_size / 2;
            size_t y = vec(1) + m_grid_size / 2;
            assert(x < m_grid_size && y < m_grid_size);
            return m_hit_map[x + y * m_grid_size];
        }
        const double &hit(const Eigen::Vector2d &vec) const {
            return hit(convert_coord(vec));
        }
        double &hit(const Eigen::Vector2i &vec) {
            return const_cast<double &>(static_cast<const GridMap &>(*this).hit(vec));
        }
        double &hit(const Eigen::Vector2d &vec) {
            return const_cast<double &>(static_cast<const GridMap &>(*this).hit(vec));
        }

        const double &miss(const Eigen::Vector2i &vec) const {
            size_t x = vec(0) + m_grid_size / 2;
            size_t y = vec(1) + m_grid_size / 2;
            assert(x < m_grid_size && y < m_grid_size);
            return m_miss_map[x + y * m_grid_size];
        }
        const double &miss(const Eigen::Vector2d &vec) const {
            return miss(convert_coord(vec));
        }
        double &miss(const Eigen::Vector2i &vec) {
            return const_cast<double &>(static_cast<const GridMap &>(*this).miss(vec));
        }
        double &miss(const Eigen::Vector2d &vec) {
            return const_cast<double &>(static_cast<const GridMap &>(*this).miss(vec));
        }

        const double &annotation(const Eigen::Vector2i &vec) const {
            size_t x = vec(0) + m_grid_size / 2;
            size_t y = vec(1) + m_grid_size / 2;
            assert(x < m_grid_size && y < m_grid_size);
            return m_annotation_map[x + y * m_grid_size];
        }
        const double &annotation(const Eigen::Vector2d &vec) const {
            return annotation(convert_coord(vec));
        }
        double &annotation(const Eigen::Vector2i &vec) {
            return const_cast<double &>(static_cast<const GridMap &>(*this).annotation(vec));
        }
        double &annotation(const Eigen::Vector2d &vec) {
            return const_cast<double &>(static_cast<const GridMap &>(*this).annotation(vec));
        }

        void draw_line(const Eigen::Vector2i &v0, const Eigen::Vector2i &v1);
        void draw_line(const Eigen::Vector2d &v0, const Eigen::Vector2d &v1);

        void push_back_robot(const Eigen::Vector2d &robot) {
            m_robot_positions.insert(convert_coord(robot));
        }
        void get_robot_positions(std::vector<Eigen::Vector2i> &positions) const;
        void get_robot_positions(std::vector<Eigen::Vector2d> &positions) const;

        Rect get_bounding_box(double threshold = 0.1) const;
        void get_point_cloud(PointCloud &pc, double threshold = 0.1) const;

    private:
        double                          m_cell_size;
        size_t                          m_grid_size;
        std::vector<double>             m_hit_map;
        std::vector<double>             m_miss_map;
        std::vector<double>             m_annotation_map;
        std::set<Eigen::Vector2i, CompareVec> m_robot_positions;

        void write_gm(const std::string &filename) const;
        void write_png(const std::string &filename) const;
};

#endif //#ifndef __grid_map_h__
