#include "imu_data.h"
#include "auto_free.h"

#include "logging/logging.h"
#include "logging/except.h"

#include <cstdio>
#include <cinttypes>
#include <cmath>

const uint64_t VERSION = 2021012201;

IMUData::IMUData() {
    m_smoothed_index = 1;
    m_rotations = 0;
}

void IMUData::read(const std::string &filename) {
    m_data.clear();

    AutoFree<FILE *, int> file(fopen(filename.c_str(), "r"), fclose);
    if (file == nullptr)
        throw FM_SYS("fopen");

    char magic[6];
    uint64_t version;
    if (fscanf(file, "%5s %" SCNu64, magic, &version) != 2)
        throw FM_SYS("fscanf");

    if (strncmp(magic, "FMIMU", 5) != 0)
        throw FM("Invalid file magic");

    if (version != VERSION)
        throw FM("Invalid file version");

    uint64_t n_datapoints;
    if (fscanf(file, "%" SCNu64, &n_datapoints) != 1)
        throw FM_SYS("fscanf");

    for (uint64_t i = 0; i < n_datapoints; i++) {
        IMUDatum datum;
        int ret = fscanf(file, "%" SCNu64 "%lf %lf %lf", &datum.timestamp,
                &datum.heading, &datum.pitch, &datum.roll);
        if (ret != 4)
            throw FM_SYS("fscanf");
        m_data.push_back(datum);
    }

    info("from", filename, "read", size(), "datapoints");
}

void IMUData::write(const std::string &filename) const {
    AutoFree<FILE *, int> file(fopen(filename.c_str(), "w"), fclose);
    if (file == nullptr)
        throw FM_SYS("fopen");

    if (fprintf(file, "FMIMU %" PRIu64 "\n", VERSION) < 0)
        throw FM_SYS("fprintf");

    if (fprintf(file, "%" PRIu64 "\n", (uint64_t)size()) < 0)
        throw FM_SYS("fprintf");

    for (const IMUDatum &datum: m_data) {
        int ret = fprintf(file, "%" PRIu64 " %lf %lf %lf\n", datum.timestamp,
                    datum.heading, datum.roll, datum.pitch);
        if (ret < 0)
            throw FM_SYS("fprintf");
    }

    info("wrote", size(), "datapoints to", filename);
}

void IMUData::push_back(IMUDatum datum) {
    if (m_data.size() > 0) {
        double heading0 = m_data.back().heading;
        double heading1 = datum.heading + 2 * M_PI * m_rotations;
        double diff = heading1 - heading0;
        if (diff < -M_PI )
            m_rotations++;
        else if (diff > M_PI)
            m_rotations--;

        datum.heading += 2 * M_PI * m_rotations;
    }

    m_data.push_back(datum);
}

void IMUData::smooth_heading() {
    for (size_t i = m_smoothed_index; i < size() - 1; i++) {
        double dt0 = m_data[i].timestamp - m_data[i - 1].timestamp;
        double dt1 = m_data[i + 1].timestamp - m_data[i].timestamp;
        double dh0 = m_data[i].heading - m_data[i - 1].heading;
        double dh1 = m_data[i + 1].heading - m_data[i].heading;
        double peak = std::min(std::abs(dh0 / dt0), std::abs(dh1 / dt1));
        if (peak > 0.01) {
            m_data[i].heading = (m_data[i - 1].heading + m_data[i + 1].heading) / 2;
            info("removed peak at:", m_data[i].timestamp);
        }
    }

    if (size() > 1)
        m_smoothed_index = size() - 1;
}

IMUDatum IMUData::interp(uint64_t timestamp, size_t &hint) const {
    const IMUDatum &before = find(timestamp, hint, true);
    const IMUDatum &after = find(timestamp, hint, false);

    if (before.timestamp == after.timestamp)
        return before;

    double tb = before.timestamp;
    double ta = after.timestamp;
    double t = timestamp;

    double alpha = (t - tb) / (ta - tb);

    IMUDatum datum;
    datum.timestamp = timestamp;
    datum.heading   = (1.0 - alpha) * before.heading + alpha * after.heading;
    datum.pitch     = (1.0 - alpha) * before.pitch   + alpha * after.pitch;
    datum.roll      = (1.0 - alpha) * before.roll    + alpha * after.roll;

    return datum;
}

/*  if ts is lower than the first element return the first element
 *  if ts is between two elements i and i + 1, return
 *      i       if before == true
 *      i + 1   if before == false
 *  if ts is greater than the last element, return the last element
 *
 *  assume there is atleast one item in the data structure
 */
const IMUDatum &IMUData::find(uint64_t timestamp, size_t &hint, bool before) const {
    assert(m_data.size() != 0);

    // edge case where timestamp <= m_data[0]
    if (timestamp <= m_data[0].timestamp) {
        hint = 0;
        return m_data[hint];
    }

    // edge case where timestamp >= m_data[-1]
    if (timestamp >= m_data.back().timestamp) {
        hint = m_data.size() - 1;
        return m_data[hint];
    }

    if (hint == HINT_UNKOWN || !find_hint(timestamp, hint)) {
        find_bin_search(timestamp, hint);
    }

    assert(hint + 1 < m_data.size());

    // add 1 of before == false
    return m_data[hint + (before ? 0 : 1)];
}

/*
 * if m_data[i] <= ts and ts < m_data[i] return i
 * else undefined
 */
bool IMUData::find_hint(uint64_t timestamp, size_t &hint) const {
    assert(m_data.size() != 0);

    for (size_t i = hint; (i + 1) < m_data.size() - 1 && i < hint + 2; i++) {
        uint64_t a = m_data[i].timestamp;
        uint64_t b = m_data[i + 1].timestamp;

        if (a <= timestamp && timestamp < b) {
            hint = i;
            return true;
        }
    }

    return false;
}

/*
 * if m_data[i] <= ts and ts < m_data[i] return i
 * else undefined
 */
void IMUData::find_bin_search(uint64_t timestamp, size_t &hint) const {
    assert(m_data.size() != 0);

    size_t l = 0;
    size_t r = m_data.size() - 1;

    while (l <= r) {
        size_t m = l + (r - l) / 2;

        if (m_data[m].timestamp == timestamp) {
            hint = m;
            return;
        }
        else if (m_data[m].timestamp < timestamp)
            l = m + 1;
        else /* if (m_data[m] > timestamp) */
            r = m - 1;
    }

    hint = r;

    assert(l - 1 == r);
    assert(m_data[r].timestamp <= timestamp && timestamp < m_data[l].timestamp);
}

