#ifndef __map_h__
#define __map_h__

#include <Eigen/Dense>

#include <vector>

template <typename T>
class Map {
    public:
        Map(double cell_size, size_t grid_size) :
            m_cell_size(cell_size),
            m_grid_size(grid_size),
            m_map(m_grid_size * m_grid_size, static_cast<T>(0)) {}

        double get_cell_size() const {
            return m_cell_size;
        }

        void fill(T value) {
            std::fill(m_map.begin(), m_map.end(), value);
        }

        Eigen::Vector2i convert_coord(const Eigen::Vector2d &vec) const {
            Eigen::Vector2d x = vec / m_cell_size;
            return Eigen::Vector2i(std::round(x(0)), std::round(x(1)));
        }
        Eigen::Vector2d convert_coord(const Eigen::Vector2i &vec) const {
            Eigen::Vector2d x(vec(0), vec(1));
            return x * m_cell_size;
        }

        size_t get_width() const {
            return m_grid_size;
        }
        size_t get_height() const {
            return m_grid_size;
        }

        bool valid_coord(const Eigen::Vector2i &vec) const {
            size_t x = vec(0) + m_grid_size / 2;
            size_t y = vec(1) + m_grid_size / 2;
            return x < m_grid_size && y < m_grid_size;
        }
        bool valid_coord(const Eigen::Vector2d &vec) const {
            return valid_coord(convert_coord(vec));
        }


        const T &operator[](const Eigen::Vector2i &vec) const {
            assert(valid_coord(vec));
            size_t x = vec(0) + m_grid_size / 2;
            size_t y = vec(1) + m_grid_size / 2;
            return m_map[x + y * m_grid_size];
        }
        const T &operator[](const Eigen::Vector2d &vec) const {
            return (*this)[convert_coord(vec)];
        }
        T &operator[](const Eigen::Vector2i &vec) {
            return const_cast<T &>(static_cast<const Map &>(*this)[vec]);
        }
        T &operator[](const Eigen::Vector2d &vec) {
            return const_cast<T &>(static_cast<const Map &>(*this)[vec]);
        }

    private:
        double m_cell_size;
        size_t m_grid_size;
        std::vector<T> m_map;
};

#endif //#ifndef __map_h__
