#include <Eigen/Dense>

#include "passes.h"

#include "logging/logging.h"

#include <set>
#include <thread>

using namespace Eigen;

namespace {

double error_func(const GridMap &gm, const PointCloud &pc) {
    double err = 0.0;
    unsigned count = 0;

    for (size_t i = 0; i < pc.size(); i++) {
        if (!gm.valid_coord(pc[i].data))
            continue;

        err += std::abs(1.0 - gm.hit(pc[i].data));
        count += 1;
    }

    return err / count;
}

void thread_helper(double *err, Matrix3d t, const GridMap *gm, const PointCloud *pc) {
    PointCloud tmp = *pc;
    tmp.transform(t);
    *err = error_func(*gm, tmp);
}

}

// TODO rename etc.
double align_pc_to_gm2(const GridMap *gm, const PointCloud *pc, Matrix3d &t) {
    const double DELTA_OFF =    2.0     / 100.0;        // cm
    const double MAX_DIF_OFF =  10.0    / 100.0;        // cm
    const double DELTA_ANG =    0.05;                   // rad
    const double MAX_DIF_ANG =  0.0;                    // rad

    PointCloud baseline = *pc;
    baseline.transform(t);
    double best_error = error_func(*gm, baseline);

    std::vector<Matrix3d> transformations;

    for (double dx = -MAX_DIF_OFF; dx <= MAX_DIF_OFF * 1.01; dx += DELTA_OFF) {
    for (double dy = -MAX_DIF_OFF; dy <= MAX_DIF_OFF * 1.01; dy += DELTA_OFF) {
    for (double th = -MAX_DIF_ANG; th <= MAX_DIF_ANG * 1.01; th += DELTA_ANG) {
        //double th = 0.0;

        Matrix3d trans;
        trans <<    std::cos(th),  -std::sin(th),   dx,
                    std::sin(th),   std::cos(th),   dy,
                    0.0,            0.0,            1.0;
        transformations.push_back(trans * t);
    }}}

    std::vector<std::thread> threads;
    std::vector<double> errors(transformations.size(), 0.0);
    for (size_t i = 0; i < transformations.size(); i++) {
        threads.push_back(std::thread(thread_helper, &errors[i], transformations[i], gm, pc));
    }

    // the global dx,dy
    double dx = t(0, 2);
    double dy = t(1, 2);
    double theta = std::atan2(t(1, 0), t(0, 0));
    for (size_t i = 0; i < transformations.size(); i++) {
        threads[i].join();
        if (errors[i] < best_error) {
            t = transformations[i];
            best_error = errors[i];
        }
    }

    // updated dx,dy
    dx = t(0,2) - dx;
    dy = t(1,2) - dy;
    theta = theta - std::atan2(t(1, 0), t(0, 0));

    info("error:", best_error, "transform: (", dx, ",", dy, ",", theta, ")");

    return best_error;
}

double pass_imu_max_heading_vel(const IMUData &imu_data, uint64_t begin, uint64_t end) {
    size_t i = HINT_UNKOWN;
    size_t j = HINT_UNKOWN;
    imu_data.find(begin, i);
    imu_data.find(end, j);
    if (j + 1 < imu_data.size())
        j++;

    double max = 0.0;
    for (size_t k = i + 1; k < j; k++) {
        double dh = imu_data[k].heading - imu_data[k - 1].heading;
        double dt = imu_data[k].timestamp - imu_data[k - 1].timestamp;
        max = std::max(std::abs(dh/dt), max);
    }

    return max;
}

double pass_imu_max_pr_vel(const IMUData &imu_data, uint64_t begin, uint64_t end) {
    size_t i = HINT_UNKOWN;
    size_t j = HINT_UNKOWN;

    imu_data.find(begin, i);
    imu_data.find(end, j);

    double max = 0.0;
    for (size_t k = i + 1; k < j; k++) {
        double dp = imu_data[k].pitch - imu_data[k - 1].pitch;
        double dr = imu_data[k].roll - imu_data[k - 1].roll;
        double dt = imu_data[k].timestamp - imu_data[k - 1].timestamp;

        max = std::max(std::abs(dp/dt), max);
        max = std::max(std::abs(dr/dt), max);
    }

    return max;
}

void pass_pc_clamp(PointCloud &pc, double min_r, double max_r) {
    Vector2d center = pc.get_center();
    PointCloud dst(center);
    for (size_t i = 0; i < pc.size(); i++) {
        double r = (pc[i].data - center).norm();
        if (min_r <= r && r <= max_r)
            dst.push_back(pc[i]);
    }

    info("removed:", pc.size() - dst.size(), "of the", pc.size(), "points");
    pc = dst;
}

void pass_pc_imu_correct(PointCloud &pc, const IMUData &imu_data) {
    size_t hint = HINT_UNKOWN;
    for (size_t i = 0; i < pc.size(); i++) {
        PointCloudDatum &datum = pc[i];
        double phi = imu_data.interp(datum.timestamp, hint).heading;

        Matrix2d rot;
        rot <<  std::cos(phi), -std::sin(phi),
                std::sin(phi),  std::cos(phi);

        datum.data = rot * datum.data;
    }
}

void pass_gm_insert_pc(GridMap &gm, const PointCloud &pc) {
    // insert hits
    for (size_t i = 0; i < pc.size(); i++) {
        if (! gm.valid_coord(pc[i].data))
            continue;

        gm.hit(pc[i].data) = std::min(1.0, gm.hit(pc[i].data) + 0.1);
        gm.miss(pc[i].data) = std::max(0.0, gm.miss(pc[i].data) - 0.05);
    }

    // insert misses
    // we use the DDA algorithm to draw lines
    Vector2i pc_center = gm.convert_coord(pc.get_center());
    for (size_t i = 0; i < pc.size(); i++) {
        Vector2i hit = gm.convert_coord(pc[i].data);
        Vector2i delta = hit - pc_center;
        unsigned steps = std::max(std::abs(delta(0)), std::abs(delta(1)));
        for (unsigned step = 0; step < steps; step++) {
            Vector2i miss = step * delta / steps + pc_center;
            if (! gm.valid_coord(miss))
                continue;

            if (hit == miss)
                continue;

            gm.hit(miss) = std::max(0.0, gm.hit(miss) - 0.05);
            gm.miss(miss) = std::min(1.0, gm.miss(miss) + 0.1);
        }
    }
}

void pass_gm_insert_pc(GridMap &gm, const PointCloud &pc, const Rect &rect) {
    // insert hits
    for (size_t i = 0; i < pc.size(); i++) {
        if (! gm.valid_coord(pc[i].data))
            continue;

        if (!rect.inside(pc[i].data))
                continue;

        gm.hit(pc[i].data) = std::min(1.0, gm.hit(pc[i].data) + 0.1);
        gm.miss(pc[i].data) = std::max(0.0, gm.miss(pc[i].data) - 0.05);
    }

    // insert misses
    // we use the DDA algorithm to draw lines
    Vector2i pc_center = gm.convert_coord(pc.get_center());
    for (size_t i = 0; i < pc.size(); i++) {
        Vector2i hit = gm.convert_coord(pc[i].data);
        Vector2i delta = hit - pc_center;
        unsigned steps = std::max(std::abs(delta(0)), std::abs(delta(1)));
        for (unsigned step = 0; step < steps; step++) {
            Vector2i miss = step * delta / steps + pc_center;
            if (! gm.valid_coord(miss))
                continue;

            if (hit == miss)
                continue;

            if (!rect.inside(gm.convert_coord(miss)))
                    continue;

            gm.hit(miss) = std::max(0.0, gm.hit(miss) - 0.05);
            gm.miss(miss) = std::min(1.0, gm.miss(miss) + 0.1);
        }
    }
}

void pass_gm_floodfill_clean(GridMap &gm, double factor, double threshold) {
    std::set<std::pair<int, int>> queue;
    std::set<std::pair<int, int>> visited;
    std::set<std::pair<int, int>> reachable;

    std::vector<Vector2i> robot_positions;
    gm.get_robot_positions(robot_positions);
    for (const Vector2i &robot: robot_positions) {
        queue.insert({robot(0), robot(1)});
    }

    while (!queue.empty()) {
        auto it = queue.begin();
        Vector2i coord(it->first, it->second);
        visited.insert(*it);
        queue.erase(it);

        if (gm.hit(coord) > threshold) {
            reachable.insert({coord(0), coord(1)});
        }
        if (gm.miss(coord) > threshold) {
            reachable.insert({coord(0), coord(1)});
            for (int i = -1; i <= 1; i++) {
            for (int j = -1; j <= 1; j++) {
                Vector2i neig = coord + Vector2i(i, j);
                if (gm.valid_coord(neig) &&
                        visited.find({neig(0), neig(1)}) == visited.end() &&
                        gm.hit(coord) <= threshold) {
                    queue.insert({neig(0), neig(1)});
                }
            }}
        }
    }

    int width = gm.get_width();
    int height = gm.get_height();
    for (int y = -height / 2; y < height / 2; y++) {
        for (int x = -width / 2; x < width / 2; x++) {
            Vector2i coord(x, y);
            gm.hit(coord)  -= factor;
            gm.miss(coord) -= factor;
        }
    }

    for (auto item: reachable) {
        Vector2i coord(item.first, item.second);

        gm.hit(coord)  += factor;
        gm.miss(coord) += factor;
    }

    for (int y = -height / 2; y < height / 2; y++) {
        for (int x = -width / 2; x < width / 2; x++) {
            Vector2i coord(x, y);

            gm.hit(coord)  = std::max(0.0, std::min(1.0, gm.hit(coord)));
            gm.miss(coord) = std::max(0.0, std::min(1.0, gm.miss(coord)));
        }
    }

    info("visited:", visited.size(), "reachable:", reachable.size());
}

