#include "point_cloud.h"
#include "auto_free.h"

#include "logging/logging.h"
#include "logging/except.h"

#include <cstdio>
#include <cinttypes>

using namespace Eigen;

const uint64_t VERSION = 2021012601;

namespace {
    inline Vector3d vec2_to_vec3(const Vector2d &vec) {
        return Vector3d(vec(0), vec(1), 1.0);
    }

    inline Vector2d vec3_to_vec2(const Vector3d &vec) {
        return Vector2d(vec(0), vec(1));
    }
}

void PointCloud::clear(Vector2d center) {
    m_center = center;
    m_data.clear();
    m_skip = false;
}

void PointCloud::read(const std::string &filename) {
    m_data.clear();
    m_center = Vector2d::Zero();

    AutoFree<FILE *, int> file(fopen(filename.c_str(), "r"), fclose);
    if (file == nullptr)
        throw FM_SYS("fopen");

    char magic[5];
    uint64_t version;
    if (fscanf(file, "%4s %" SCNu64, magic, &version) != 2)
        throw FM_SYS("fscanf");

    if (strncmp(magic, "FMPC", 4) != 0)
        throw FM("Invalid file magic");

    if (version != VERSION)
        throw FM("Invalid file version");

    uint64_t t0, t1;
    if (fscanf(file, "%" SCNu64 " %" SCNu64, &t0, &t1) != 2)
        throw FM_SYS("fscanf");

    uint64_t n_datapoints;
    if (fscanf(file, "%" SCNu64, &n_datapoints) != 1)
        throw FM_SYS("fscanf");

    for (uint64_t i = 0; i < n_datapoints; i++) {
        double x, y;
        if (fscanf(file, "%lf %lf", &x, &y) != 2)
            throw FM_SYS("fscanf");
        push_back({0, Vector2d(x, y)});
    }

    calc_timestamps(t0, t1);

    info("from", filename, "read", size(), "datapoints");
}

void PointCloud::calc_timestamps(uint64_t t0, uint64_t t1) {
    std::vector<double> headings;
    headings.reserve(m_data.size());

    for (size_t i = 0; i < m_data.size(); i++) {
        Vector2d vec = m_data[i].data;
        double heading = std::atan2(vec(1), vec(0));
        headings.push_back(heading);
    }

    int n = 0;
    for (size_t i = 1; i < m_data.size(); i++) {
        double diff = (headings[i] + n * 2 * M_PI) - headings[i - 1];
        if (diff < -M_PI)
            n++;
        // no need to add the other case since our LIDAR rotates CW (negative
        // rotation angle).

        headings[i] += n * 2 * M_PI;
    }

    double h0 = headings.front();
    double h1 = headings.back();

    for (size_t i = 0; i < m_data.size(); i++) {
        double alpha = (headings[i] - h0) / (h1 - h0);
        m_data[i].timestamp = t0 + (t1 - t0) * alpha;
    }
}

void PointCloud::write(const std::string &filename) const {
    assert(m_data.size() != 0);

    AutoFree<FILE *, int> file(fopen(filename.c_str(), "w"), fclose);
    if (file == nullptr)
        throw FM_SYS("fopen");

    if (fprintf(file, "FMPC %" PRIu64 "\n", VERSION) < 0)
        throw FM_SYS("fprintf");

    uint64_t t0 = m_data.front().timestamp;
    uint64_t t1 = m_data.back().timestamp;
    if (fprintf(file, "%" PRIu64 " %" PRIu64 "\n", t0, t1) < 0)
        throw FM_SYS("fprintf");

    if (fprintf(file, "%" PRIu64 "\n", (uint64_t)size()) < 0)
        throw FM_SYS("fprintf");

    for (const PointCloudDatum &x: m_data) {
        const Vector2d &vec = x.data;
        if (fprintf(file, "%lf %lf\n", vec(0), vec(1)) < 0)
            throw FM_SYS("fprintf");
    }

    if (fflush(file) != 0)
        throw FM_SYS("fflush");

    info("wrote", size(), "datapoints to", filename);
}

void PointCloud::transform(const Eigen::Matrix3d &t) {
    m_center = vec3_to_vec2(t * vec2_to_vec3(m_center));
    for (size_t i = 0; i < m_data.size(); i++) {
        m_data[i].data = vec3_to_vec2(t * vec2_to_vec3(m_data[i].data));
    }
}

void PointCloud::set_skip(bool skip) {
    m_skip = skip;
}

bool PointCloud::get_skip() const {
    return m_skip;
}

