#include "battery.h"
#include "misc.h"

#include <iostream>

namespace {

const double POLLING_RATE       = 10;       // Hz
const double REPORT_RATE        = 1;        // Hz
const double VOLTAGE_FULL       = 12.66;    // V
const double VOLTAGE_EMPTY      = 8.0;      // V
const double BROWNOUT_THRESHOLD = 0.7;

const std::chrono::milliseconds POLLING_TIME(static_cast<int>(1000 / POLLING_RATE));

}

Battery::Battery() {
    m_voltage = -1.0;
    m_voltage_reported = true;
    m_brownout = false;
    m_brownout_reported = true;
    m_sum = 0.0;
    m_count = 0;
    m_next_push = std::chrono::high_resolution_clock::now();
}

bool Battery::should_push() {
    auto t = std::chrono::high_resolution_clock::now();
    return t > m_next_push;
}

void Battery::push_voltage(double voltage) {
    if (! should_push())
        return;

    if (voltage < 0.0 || voltage > 30.0) {
        std::cerr << "Warning: unrealistic voltage: " << voltage << " V" << std::endl;
        return;
    }

    m_sum += voltage;
    m_count++;
    m_next_push = std::chrono::high_resolution_clock::now() + POLLING_TIME;

    if (m_count >= POLLING_RATE / REPORT_RATE) {
        m_voltage = m_sum / m_count;
        m_voltage_reported = false;
        m_brownout = false;
        m_brownout_reported = false;

        m_sum = 0;
        m_count = 0;
    }

    if (voltage < BROWNOUT_THRESHOLD * m_voltage)
        m_brownout = true;
}

double Battery::get_voltage() {
    return m_voltage;
}

bool Battery::get_brownout() {
    if (m_brownout && !m_brownout_reported) {
        m_brownout_reported = true;
        return true;
    }

    return false;
}

bool Battery::get_level(double &level) {
    level = (m_voltage - VOLTAGE_EMPTY) / (VOLTAGE_FULL - VOLTAGE_EMPTY);
    level = clamp(level, 0.0, 1.0);

    if (!m_voltage_reported) {
        m_voltage_reported = true;
        return true;
    }
    return false;
}

