#include "bno055.h"

#include <cstdint>
#include <cerrno>
#include <cstring>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/i2c-dev.h>
#if __has_include(<i2c/smbus.h>)
#include <i2c/smbus.h>
#endif
#include <unistd.h>
#include <iomanip>
#include <cmath>
#include <cassert>

namespace {
// Register definition
const uint8_t CHIP_ID =             0x00;
const uint8_t ACC_ID =              0x01;
const uint8_t MAG_ID =              0x02;
const uint8_t GYR_ID =              0x03;
const uint8_t PAGE_ID =             0x07;
const uint8_t LIA_DATA_X_LSB =      0x28;
const uint8_t EUL_HEADING_LSB =     0x1a;
const uint8_t CALIB_STAT =          0x35;
const uint8_t SYS_STATUS =          0x39;
const uint8_t OPR_MODE =            0x3d;
const uint8_t PWR_MODE =            0x3e;
const uint8_t SYS_TRIGGER =         0x3f;
const uint8_t QUA_Data_W_LSB =      0x20;

// Constants definition
const uint8_t RST_SYS =            0x20;
}

BNO055::BNO055(const char *i2c_dev, uint32_t addr) {
    m_fd = open(i2c_dev, O_RDWR, 0);
    if (m_fd == -1)
        throw BNO055Exception("Failed to open i2c device");

    if (ioctl(m_fd, I2C_SLAVE, addr) == -1)
    {
        close(m_fd);
        throw BNO055Exception("Failed to open set slave address");
    }

    m_pitch_off = 0.0;
    m_roll_off = 0.0;

    init();
}

BNO055::~BNO055() {
    try {
        reset();
    }
    catch (...) {
        // if this fails, it fails
    }

    close(m_fd);
}

void BNO055::set_offset(double pitch_off, double roll_off) {
    m_pitch_off = pitch_off;
    m_roll_off = roll_off;
}

bool BNO055::ready() {
    uint8_t status = i2c_recv(SYS_STATUS);

    // the device is ready when it is in FUSION MODE
    return status == 5;
}

bool BNO055::calibrated() {
    uint8_t cal = i2c_recv(CALIB_STAT);

    int __attribute__((unused)) mag = cal & 0x03;
    int __attribute__((unused)) acc = (cal >> 2) & 0x03;
    int __attribute__((unused)) gyr = (cal >> 4) & 0x03;
    int sys = (cal >> 6);

    return sys == 3;
}

void BNO055::get(BNO055Event &event) {
    int16_t qua[4];
    int16_t lia[3];

    i2c_recv_vec(QUA_Data_W_LSB, qua, sizeof(qua));
    i2c_recv_vec(LIA_DATA_X_LSB, lia, sizeof(lia));

    for (int i = 0; i < 3; i++)
        event.lia[i] = static_cast<double>(lia[i]) / 100.0;

    double q[4];
    for (int i = 0; i < 4; i++)
        q[i] = static_cast<double>(qua[i]) / 16384.0;

    // taken from:
    // https://en.wikipedia.org/wiki/Conversion_between_quaternions_and_Euler_angles

    // yaw (z-axis rotation)
    double siny_cosp = 2 * (q[0] * q[3] + q[1] * q[2]);
    double cosy_cosp = 1.0 - 2 * (q[2] * q[2] + q[3] * q[3]);
    event.eul[0] = -180.0 * std::atan2(siny_cosp, cosy_cosp) / M_PI;
    if (event.eul[0] < 0.0)
        event.eul[0] += 360.0;

    // pitch (y-axis rotation)
    double sinp = 2 * (q[0] * q[2] - q[3] * q[0]);
    if (std::abs(sinp) >= 1)
        event.eul[1] = std::copysign(90.0, sinp); // use 90 degrees if out of range
    else
        event.eul[1] = 180.0 * std::asin(sinp) / M_PI;

    //roll (x-axis rotation)
    double sinr_cosp = 2 * (q[0] * q[1] + q[2] * q[3]);
    double cosr_cosp = 1.0 - 2 * (q[1] * q[1] + q[2] * q[2]);
    event.eul[2] = 180.0 * std::atan2(sinr_cosp, cosr_cosp) / M_PI;

    event.eul[1] -= m_roll_off;
    event.eul[2] -= m_pitch_off;
}

void BNO055::i2c_send(uint8_t reg, uint8_t data) {
    if (i2c_smbus_write_byte_data(m_fd, reg, data) == -1)
        throw BNO055Exception("i2c bus error");
}

uint8_t BNO055::i2c_recv(uint8_t reg) {
    int32_t data = i2c_smbus_read_byte_data(m_fd, reg);
    if (data == -1)
        throw BNO055Exception("i2c bus error");

    return data;
}

void BNO055::i2c_recv_vec(uint8_t reg_start, int16_t *data, size_t len) {
    uint8_t buf[32];

    assert(sizeof(buf) >= len);

    if (i2c_smbus_read_i2c_block_data(m_fd, reg_start, len, buf) == -1) {
        throw BNO055Exception("read block bus error");
    }

    // convert to the right endianness
    for (size_t i = 0; i < len / sizeof(data[0]); i++) {
        data[i] = buf[2 * i] | (uint16_t)buf[2 * i + 1] << 8;
    }
}

void BNO055::delay(uint32_t msec) {
    usleep(msec * 1000);
}

void BNO055::init() {
    uint8_t page_id = i2c_recv(PAGE_ID);
    if (page_id != 0)
        i2c_send(PAGE_ID, 0x00);

    uint8_t chip_id = i2c_recv(CHIP_ID);
    if (chip_id != 0xa0)
        throw BNO055Exception("Connected device is not a BNO055");

    // set the normal power mode
    i2c_send(PWR_MODE, 0x00);
    delay(20);

    reset();

    // set the normal power mode
    i2c_send(PWR_MODE, 0x00);
    delay(20);

    // set the operation mode to NDOF
    i2c_send(OPR_MODE, 0x0c);
    delay(10);
}

void BNO055::reset() {
    // switch to config mode
    i2c_send(OPR_MODE, 0x00);
    delay(20);

    // reset the device and wait for it to reboot
    i2c_send(SYS_TRIGGER, RST_SYS);
    delay(1000);
}
