#include "network/network.h"
#include "network/network_types.h"
#include "battery.h"
#include "bno055.h"
#include "heartbeat.h"
#include "misc.h"
#include "rpcamera.h"
#include "rplidar.h"
#include "thunder_borg.h"


#include <csignal>
#include <cstring>
#include <ctime>
#include <iostream>
#include <mutex>
#include <sys/types.h>
#include <sys/wait.h>
#include <thread>
#include <unistd.h>

bool        g_quit = false;
uint64_t    g_conn_epoch;

void video_thread(RPCamera *camera, Network *net, std::mutex *net_lock) {
    std::cerr << "INFO: video thread started" << std::endl;

    while (! g_quit) {
        char buf[2048];
        size_t size = sizeof(buf);

        try {
            camera->get_data(buf, size);
        }
        catch (std::exception &e) {
            std::cerr << "camera->get_data() failed: " << e.what() << std::endl;
            g_quit = true;
            return;
        }

        net_lock->lock();
        net->send_pkt(NETWORK_TYPE_VIDEO_STREAM, buf, size);
        net_lock->unlock();
    }
}

void lidar_thread(RPLidar *lidar, Network *net, std::mutex *net_lock) {
    while (! g_quit) {
        // obtain a lidar measurement
        RPLidarPoint points[RP_LIDAR_MAX_POINTS];
        size_t size = RP_LIDAR_MAX_POINTS;
        try {
            lidar->get_points(points, size);
        }
        catch (std::exception &e) {
            std::cerr << "lidar_thread: " << e.what() << std::endl;
            g_quit = true;
            break;
        }

        // send the lidar data
        uint8_t buf[4 + 4 * RP_LIDAR_MAX_POINTS];
        uint32_t timestamp = htonl((utime() - g_conn_epoch) / 1000);
        memcpy(buf, &timestamp, 4);

        uint16_t *network_buffer = (uint16_t *)&buf[4];
        size_t network_count = 0;
        for (size_t i = 0; i < size; i++) {
            if (points[i].distance > 50) { // 50mm
                network_buffer[2 * network_count] = htons(points[i].angle);
                network_buffer[2 * network_count + 1] = htons(points[i].distance);
                network_count++;
            }
        }

        if (size > 0) {
            net_lock->lock();
            net->send_pkt(NETWORK_TYPE_LIDAR_POINTS, buf, 4 + 4 * network_count);
            net_lock->unlock();
        }
    }
}

void imu_thread(BNO055 *imu, Network *net, std::mutex *net_lock, std::mutex *i2c_lock) {
    const double FREQ = 100.0; // Hz

    const size_t PKT_SIZE = 14; // size of a single packt
    const int N = 10;           // amount of packets to pack before sending

    int frame_n = 0;
    uint8_t buf[N * PKT_SIZE];
    while (!g_quit) {
        uint64_t start_time = utime();

        // obtain a IMU measurment
        BNO055Event ev;
        i2c_lock->lock();
        imu->get(ev);
        i2c_lock->unlock();

        // send the IMU data
        uint16_t heading    = htons(ev.eul[0]           / 360.0 * UINT16_MAX);
        uint16_t roll       = htons((ev.eul[1] + 90.0)  / 360.0 * UINT16_MAX);
        uint16_t pitch      = htons((ev.eul[2] + 180.0) / 360.0 * UINT16_MAX);
        uint8_t status      = (imu->ready()      ? 0x01 : 0x00) |
                              (imu->calibrated() ? 0x02 : 0x00);
        int8_t acc[3];
        for (int i = 0; i < 3; i++)
            acc[i] = clamp(ev.lia[i] / 2.0, -1.0, 1.0) * INT8_MAX;
        uint32_t timestamp = htonl((utime() - g_conn_epoch) / 1000);

        uint8_t *imu_buffer = buf + PKT_SIZE * frame_n;
        memcpy(imu_buffer,      &heading,   2);
        memcpy(imu_buffer + 2,  &roll,      2);
        memcpy(imu_buffer + 4,  &pitch,     2);
        memcpy(imu_buffer + 6,  &status,    1);
        memcpy(imu_buffer + 7,  acc,        3);
        memcpy(imu_buffer + 10, &timestamp, 4);
        frame_n++;

        if (frame_n == N) {
            net_lock->lock();
            net->send_pkt(NETWORK_TYPE_IMU, buf, N * PKT_SIZE);
            net_lock->unlock();

            frame_n = 0;
        }

        uint64_t end_time = utime();
        int64_t wait_time = (1000000 / FREQ) - (end_time - start_time);
        if (wait_time > 0)
            usleep(wait_time);
    }
}

void sig_handler(int signo) {
    (void)signo;

    std::cout << "Quitting ..." << std::endl;
    g_quit = true;
}

void set_signal(int signo, void (*func)(int), bool restart) {
    struct sigaction sa;
    sa.sa_handler = func;
    sa.sa_flags = 0;
    if (restart)
        sa.sa_flags |= SA_RESTART;
    sigemptyset(&sa.sa_mask);

    if (sigaction(signo, &sa, NULL) == -1)
        throw std::system_error(errno, std::generic_category());
}

int start() {
    RPLidar lidar("/dev/ttyUSB0");
    RPCamera camera(960, 540, 20);
    ThunderBorg controller("/dev/i2c-1", 0x15);
    BNO055 imu("/dev/i2c-1");
    Battery battery;

    set_signal(SIGINT, sig_handler, false);
    set_signal(SIGTERM, sig_handler, false);

    std::cerr << "Waiting for connection ... " << std::flush;
    Network net("0.0.0.0", 3142, true);
    std::cerr << "Connected" << std::endl;

    g_conn_epoch = utime();

    std::mutex net_lock;
    std::mutex i2c_lock;

    set_signal(SIGINT, sig_handler, true);
    set_signal(SIGTERM, sig_handler, true);
    set_signal(SIGPIPE, SIG_IGN, true);

    std::thread vthread(video_thread, &camera, &net, &net_lock);
    std::thread lthread(lidar_thread, &lidar, &net, &net_lock);
    std::thread ithread(imu_thread, &imu, &net, &net_lock, &i2c_lock);

    Heartbeat hb;

    while (! g_quit) {
        net_lock.lock();
        g_quit = !net.connected();
        net_lock.unlock();

        for (;;) {
            bool recved_pkt;
            uint8_t pkt_type;
            char payload[NETWORK_MAX_PAYLOAD_SIZE];
            size_t payload_size = NETWORK_MAX_PAYLOAD_SIZE;

            net_lock.lock();
            recved_pkt = net.recv_pkt(pkt_type, payload, payload_size);
            net_lock.unlock();
            if (! recved_pkt)
                break;

            switch (pkt_type) {
                case NETWORK_TYPE_MOTOR:
                    if (hb.alive()) {
                        i2c_lock.lock();
                        controller.set_motor_a(((int8_t)payload[1]) / 127.0);
                        controller.set_motor_b(((int8_t)payload[0]) / 127.0);
                        i2c_lock.unlock();
                    }
                    break;

                case NETWORK_TYPE_HEARTBEAT:
                    // no need to do anything, as we considder each message as a HB
                    break;

                case NETWORK_TYPE_TERMINATE:
                    g_quit = true;
                    break;

                default:
                    break;
            }

            hb.beat();
        }

        if (hb.should_stop()) {
            std::cerr << "Warning HB timeout, stopping motors" << std::endl;
            i2c_lock.lock();
            controller.set_motors(0.0);
            i2c_lock.unlock();
        }

        if (battery.should_push()) {
            i2c_lock.lock();
            battery.push_voltage(controller.get_battery_voltage());
            i2c_lock.unlock();
        }

        if (battery.get_brownout()) {
            std::cerr << "Warning: brownout detected" << std::endl;

            net_lock.lock();
            net.send_pkt(NETWORK_TYPE_BAT_BROWNOUT, NULL, 0);
            net_lock.unlock();
        }

        double level;
        if (battery.get_level(level)) {
            uint8_t data = UINT8_MAX * level;

            net_lock.lock();
            net.send_pkt(NETWORK_TYPE_BAT_LEVEL, &data, 1);
            net_lock.unlock();
        }

        usleep(10000);
    }

    ithread.join();
    lthread.join();
    vthread.join();

    return 0;
}

int main() {
    int ret = 1;
    try {
        ret = start();
        std::cerr << "Server terminated" << std::endl;
    }
    catch (std::exception &e) {
        std::cerr << "Server terminated unexpected with error: '"
            << e.what() << "'" << std::endl;
    }

    return ret;
}
