#include <cassert>
#include <csignal>
#include <cstring>
#include <unistd.h>
#include <sys/wait.h>
#include <iostream>

#include "rpcamera.h"

RPCamera::RPCamera(int width, int height, int fps) {
    if (width < 1 || width > 3280)
        throw RPCameraException("Invalid with specified");
    if (height < 1 || height > 2464)
        throw RPCameraException("Invalid height specified");
    if (fps < 1 || fps > 200)
        throw RPCameraException("Invalid fps specified");

    m_raspivid = -1;
    m_pipeout = -1;

    char width_str[16];
    char height_str[16];
    char fps_str[16];
    sprintf(width_str, "%d", width);
    sprintf(height_str, "%d", height);
    sprintf(fps_str, "%d", fps);

    int fds[2];
    if (pipe(fds) == -1)
        throw std::system_error(errno, std::generic_category());

    m_raspivid = fork();
    if (m_raspivid == -1) {
        close(fds[0]);
        close(fds[1]);

        throw std::system_error(errno, std::generic_category());
    }

    if (m_raspivid == 0) {
        // child
        close(STDOUT_FILENO);
        close(fds[0]);

        if (dup2(fds[1], STDOUT_FILENO) == -1)
        {
            std::cerr << "dup2() Failed: " << strerror(errno) << std::endl;
            abort();
        }

        if(execl("/usr/bin/raspivid", "raspivid", "-t", "0", "-vf", "-hf", "-fl",
                "-w", width_str, "-h", height_str, "-fps", fps_str, "-o", "-", (char *)NULL) == -1) {
            std::cerr << "execl() failed: " << strerror(errno) << std::endl;
            abort();
        }
    }
    else {
        // parent
        close(fds[1]);
        m_pipeout = fds[0];
    }

    // wait .5s to check if the child spawned correctly, this is useful
    // because it prevents other components from initializing just for the
    // server to crash on the next get_data() call.
    usleep(500000);
    if (! raspivid_alive()) {
        close(m_pipeout);
        throw RPCameraException("raspivid process already died");
    }
}

RPCamera::~RPCamera() {
    if (raspivid_alive()) {
        kill(m_raspivid, SIGKILL);
        waitpid(m_raspivid, NULL, 0);
    }

    close(m_pipeout);
}

void RPCamera::get_data(void *buf, size_t &size) {
    if (! raspivid_alive())
        throw RPCameraException("raspivid process died");

    ssize_t ret = read(m_pipeout, buf, size);
    if (ret == -1)
        throw std::system_error(errno, std::generic_category());
    else if (ret == 0)
        throw RPCameraException("Empty response from raspivid");

    size = ret;
}

bool RPCamera::raspivid_alive() {
    if (m_raspivid == -1)
        return false;

    pid_t ret = waitpid(m_raspivid, NULL, WNOHANG);
    if (ret == -1)
        throw std::system_error(errno, std::generic_category());
    else if (ret == 0)
        return true;

    assert(ret == m_raspivid);

    // raspivid terminated
    m_raspivid = -1;
    return false;
}

