#include "thunder_borg.h"
#include "misc.h"

#include <arpa/inet.h>
#include <cerrno>
#include <cstdint>
#include <cstring>
#include <fcntl.h>
#include <iomanip>
#include <iostream>
#include <linux/i2c-dev.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <unistd.h>

#define sys_panic() ({fprintf(stderr, "%d failed %s\n", __LINE__, strerror(errno)); abort();})
#define panic(msg) ({fprintf(stderr, "%d failed %s\n", __LINE__, (msg)); abort();})

namespace {

// the minimum pwm value, where the motos just start turning
const int MOTOR_DEADZONE = 50;
// the normal speed that is considderd 0
const double NORM_SPEED_EPS = 0.01;

// ADC parameters
const int ANALOG_MAX =  0x3ff;
const double VOLTAGE_REF = 3.3; // V
const double VOLTAGE_MAX = 11 * VOLTAGE_REF; // V

const int DATA_MAX_SIZE = 6;
const int NUM_RETRIES = 3;

const uint8_t COMMAND_SET_LED1            = 1;
const uint8_t COMMAND_GET_LED1            = 2;
const uint8_t COMMAND_SET_LED2            = 3;
const uint8_t COMMAND_GET_LED2            = 4;
const uint8_t COMMAND_SET_LEDS            = 5;
const uint8_t COMMAND_SET_LED_BATT_MON    = 6;
const uint8_t COMMAND_GET_LED_BATT_MON    = 7;
const uint8_t COMMAND_SET_A_FWD           = 8;
const uint8_t COMMAND_SET_A_REV           = 9;
const uint8_t COMMAND_GET_A               = 10;
const uint8_t COMMAND_SET_B_FWD           = 11;
const uint8_t COMMAND_SET_B_REV           = 12;
const uint8_t COMMAND_GET_B               = 13;
const uint8_t COMMAND_ALL_OFF             = 14;
const uint8_t COMMAND_GET_DRIVE_A_FAULT   = 15;
const uint8_t COMMAND_GET_DRIVE_B_FAULT   = 16;
const uint8_t COMMAND_SET_ALL_FWD         = 17;
const uint8_t COMMAND_SET_ALL_REV         = 18;
const uint8_t COMMAND_SET_FAILSAFE        = 19;
const uint8_t COMMAND_GET_FAILSAFE        = 20;
const uint8_t COMMAND_GET_BATT_VOLT       = 21;
const uint8_t COMMAND_SET_BATT_LIMITS     = 22;
const uint8_t COMMAND_GET_BATT_LIMITS     = 23;
const uint8_t COMMAND_WRITE_EXTERNAL_LED  = 24;
const uint8_t COMMAND_GET_ID              = 0x99;
const uint8_t COMMAND_SET_I2C_ADD         = 0xAA;

uint8_t norm2pwm(double speed) {
    /*
    if (speed < 0)
        speed = -speed;

    double a0 = MOTOR_DEADZONE / NORM_SPEED_EPS;
    double a1 = (255.0 - MOTOR_DEADZONE) / 1.0;

    int pwm0 = a0 * speed + 0.5;
    int pwm1 = a1 * speed + MOTOR_DEADZONE + 0.5;
    return clamp(std::min(pwm0, pwm1), 0, 255);
    */

    speed = std::abs(speed);
    int pwm = (255 - MOTOR_DEADZONE) * speed + MOTOR_DEADZONE + 0.5;
    return clamp(pwm, 0, 255);
}

}

ThunderBorg::ThunderBorg(const char *i2c_dev, uint32_t addr) {
    m_fd = open(i2c_dev, O_RDWR, 0);
    if (m_fd == -1)
        sys_panic();

    if (ioctl(m_fd, I2C_SLAVE, addr) == -1)
        sys_panic();

    uint8_t data = i2c_recv_u8(COMMAND_GET_ID);
    if (data != 0x15)
        panic("Device is not a Thunder Borg");

    stop_all();
}

ThunderBorg::~ThunderBorg() {
    stop_all();
    close(m_fd);
}

void ThunderBorg::set_motors(double speed) {
    speed = clamp(speed, -1.0, 1.0);

    uint8_t pwm_speed = norm2pwm(speed);
    uint8_t command;
    if (speed > 0)
        command = COMMAND_SET_ALL_FWD;
    else
        command = COMMAND_SET_ALL_REV;

    i2c_send_u8(command, pwm_speed);
}

void ThunderBorg::set_motor_a(double speed) {
    speed = clamp(speed, -1.0, 1.0);

    uint8_t pwm_speed = norm2pwm(speed);
    uint8_t command;
    if (speed > 0)
        command = COMMAND_SET_A_FWD;
    else
        command = COMMAND_SET_A_REV;

    i2c_send_u8(command, pwm_speed);
}

void ThunderBorg::set_motor_b(double speed) {
    speed = clamp(speed, -1.0, 1.0);

    uint8_t pwm_speed = norm2pwm(speed);
    uint8_t command;
    if (speed > 0)
        command = COMMAND_SET_B_FWD;
    else
        command = COMMAND_SET_B_REV;

    i2c_send_u8(command, pwm_speed);
}

double ThunderBorg::get_battery_voltage() {
    uint16_t data = i2c_recv_u16(COMMAND_GET_BATT_VOLT);

    uint16_t raw = ntohs(data);
    double level = static_cast<double>(raw) / ANALOG_MAX;

    return VOLTAGE_MAX * level;
}

void ThunderBorg::i2c_send(uint8_t command) {
    ssize_t ret = write(m_fd, &command, sizeof(command));
    if (ret == -1)
        sys_panic();
}

void ThunderBorg::i2c_send_u8(uint8_t command, uint8_t data) {
    uint8_t buf[2] = {command, data};

    ssize_t ret = write(m_fd, buf, sizeof(buf));
    if (ret == -1)
        sys_panic();

    if (ret != 2)
        panic("i2c partial send");
}

void ThunderBorg::i2c_recv(uint8_t command, uint8_t *buf, size_t size) {
    i2c_send(command);

    ssize_t ret = read(m_fd, buf, size);
    if (ret == -1)
        sys_panic();
    if ((size_t)ret != size)
        panic("i2c partial read");
}

uint8_t ThunderBorg::i2c_recv_u8(uint8_t command) {
    if (command == 0x00)
        panic("0x00 command specified");

    uint8_t buf[DATA_MAX_SIZE] = {};
    for (int i = 0; i < NUM_RETRIES && buf[0] != command; i++)
        i2c_recv(command, buf, sizeof(buf));

    if (buf[0] != command)
        panic("i2c bus error");

    uint8_t data;
    memcpy(&data, buf + 1, sizeof(data));
    return data;
}

uint16_t ThunderBorg::i2c_recv_u16(uint8_t command) {
    if (command == 0x00)
        panic("0x00 command specified");

    uint8_t buf[DATA_MAX_SIZE] = {};
    for (int i = 0; i < NUM_RETRIES && buf[0] != command; i++)
        i2c_recv(command, buf, sizeof(buf));

    if (buf[0] != command)
        panic("i2c bus error");

    uint16_t data;
    memcpy(&data, buf + 1, sizeof(data));
    return data;
}

void ThunderBorg::stop_all() {
    i2c_send(COMMAND_ALL_OFF);
}
